import warnings
import logging
import inspect
import re
import itertools
import collections
from .connection import conn
from .diagram import Diagram, _get_tier
from .settings import config
from .errors import DataJointError, AccessError
from .jobs import JobTable
from .external import ExternalMapping
from .heading import Heading
from .utils import user_choice, to_camel_case
from .user_tables import Computed, Imported, Manual, Lookup
from .part_table import Part
from .auto_tables import AutoComputed, AutoImported
from .settings_table import Settingstable
from .table import lookup_class_name, Log, FreeTable
import types

logger = logging.getLogger(__name__)


def ordered_dir(class_):
    """
    List (most) attributes of the class including inherited ones, similar to `dir` build-in function,
    but respects order of attribute declaration as much as possible.
    This becomes unnecessary in Python 3.6+ as dicts became ordered.
    :param class_: class to list members for
    :return: a list of attributes declared in class_ and its superclasses
    """
    attr_list = list()
    for c in reversed(class_.mro()):
        attr_list.extend(e for e in (
            c._ordered_class_members if hasattr(c, '_ordered_class_members') else c.__dict__)
            if e not in attr_list)
    return attr_list


class Schema:
    """
    A schema object is a decorator for UserTable classes that binds them to their database.
    It also specifies the namespace `context` in which other UserTable classes are defined.
    """

    def __init__(self, schema_name=None, context=None, *, connection=None, create_schema=True,
                 create_tables=True, add_objects=None):
        """
        Associate database schema `schema_name`. If the schema does not exist, attempt to
        create it on the server.

        If the schema_name is omitted, then schema.activate(..) must be called later
        to associate with the database.

        :param schema_name: the database schema to associate.
        :param context: dictionary for looking up foreign key references, leave None to use local context.
        :param connection: Connection object. Defaults to datajoint.conn().
        :param create_schema: When False, do not create the schema and raise an error if missing.
        :param create_tables: When False, do not create tables and raise errors when accessing missing tables.
        :param add_objects: a mapping with additional objects to make available to the context in which table classes
        are declared.
        """
        self._log = None
        self.connection = connection
        self.database = None
        self.context = context
        self.create_schema = create_schema
        self.create_tables = create_tables
        self._jobs = None
        self.external = ExternalMapping(self)
        self.add_objects = add_objects
        self.declare_list = []
        if schema_name:
            self.activate(schema_name)

    def is_activated(self):
        return self.database is not None

    def activate(self, schema_name=None, *, connection=None, create_schema=None,
                 create_tables=None, add_objects=None):
        """
        Associate database schema `schema_name`. If the schema does not exist, attempt to
        create it on the server.
        :param schema_name: the database schema to associate.
            schema_name=None is used to assert that the schema has already been activated.
        :param connection: Connection object. Defaults to datajoint.conn().
        :param create_schema: If False, do not create the schema and raise an error if missing.
        :param create_tables: If False, do not create tables and raise errors when attempting
            to access missing tables.
        :param add_objects: a mapping with additional objects to make available to the context
            in which table classes are declared.
        """
        if schema_name is None:
            if self.exists:
                return
            raise DataJointError("Please provide a schema_name to activate the schema.")
        if self.database is not None and self.exists:
            if self.database == schema_name:  # already activated
                return
            raise DataJointError(
                "The schema is already activated for schema {db}.".format(db=self.database))
        if connection is not None:
            self.connection = connection
        if self.connection is None:
            self.connection = conn()
        self.database = schema_name
        if create_schema is not None:
            self.create_schema = create_schema
        if create_tables is not None:
            self.create_tables = create_tables
        if add_objects:
            self.add_objects = add_objects
        if not self.exists:
            if not self.create_schema or not self.database:
                raise DataJointError(
                    "Database `{name}` has not yet been declared. "
                    "Set argument create_schema=True to create it.".format(name=schema_name))
            # create database
            logger.info("Creating schema `{name}`.".format(name=schema_name))
            try:
                self.connection.query("CREATE DATABASE `{name}`".format(name=schema_name))
            except AccessError:
                raise DataJointError(
                    "Schema `{name}` does not exist and could not be created. "
                    "Check permissions.".format(name=schema_name))
            else:
                self.log('created')
        self.connection.register(self)

        # decorate all tables already decorated
        for cls, context in self.declare_list:
            if self.add_objects:
                context = dict(context, **self.add_objects)
            self._decorate_master(cls, context)

    def _assert_exists(self, message=None):
        if not self.exists:
            raise DataJointError(
                message or "Schema `{db}` has not been created.".format(db=self.database))

    def __call__(self, cls, *, context=None):
        """
        Binds the supplied class to a schema. This is intended to be used as a decorator.
        :param cls: class to decorate.
        :param context: supplied when called from spawn_missing_classes
        """
        context = context or self.context or inspect.currentframe().f_back.f_locals

        # update backup context with context
        backup = config['backup_context']
        context = {**backup, **context}

        if issubclass(cls, Part):
            raise DataJointError('The schema decorator should not be applied to Part relations')
        # automaker specific
        is_automaker = issubclass(
            cls, (AutoComputed, AutoImported, Settingstable))

        if not config['enable_automakers'] and is_automaker:
            raise DataJointError(
                'Enable automakers in config to use '
                'AutoComputed and AutoImported tables.'
            )

        elif not config['enable_python_native_blobs'] and is_automaker:
            raise DataJointError(
                'Enable native python blobs to use'
                'Autocomputed and AutoImported tables.'
            )

        if issubclass(cls, (AutoComputed, AutoImported)):
            settings_schema = (
                self if cls.settings_table._assigned_schema is None
                else cls.settings_table._assigned_schema
            )
            context[cls.settings_table._save_name] = settings_schema(
                cls.settings_table, context=context
            )
            cls.set_true_definition()
        # normal datajoint continues
        if self.is_activated():
            self._decorate_master(cls, context)
        else:
            self.declare_list.append((cls, context))
        return cls

    def _decorate_master(self, cls, context):
        """
        :param cls: the master class to process
        :param context: the class' declaration context
        """
        self._decorate_table(cls, context=dict(context, self=cls, **{cls.name: cls}))
        # Process part tables
        for part in ordered_dir(cls):
            if part[0].isupper():
                part = getattr(cls, part)
                if inspect.isclass(part) and issubclass(part, Part):
                    part._master = cls
                    # allow addressing master by name or keyword 'master'
                    self._decorate_table(part, context=dict(
                        context, master=cls, self=part, **{cls.name: cls}))

    def _decorate_table(self, table_class, context, assert_declared=False):
        """
        assign schema properties to the table class and declare the table
        """
        table_class.database = self.database
        table_class._connection = self.connection
        table_class._heading = Heading(table_info=dict(
            conn=self.connection,
            database=self.database,
            table_name=table_class.table_name,
            context=context))
        table_class._support = [table_class.full_table_name]
        table_class.declaration_context = context

        # instantiate the class, declare the table if not already
        instance = table_class()
        is_declared = instance.is_declared
        if not is_declared:
            if not self.create_tables or assert_declared:
                raise DataJointError('Table `%s` not declared' % instance.table_name)
            instance.declare(context)
        is_declared = is_declared or instance.is_declared

        # add table definition to the doc string
        if isinstance(table_class.definition, str):
            table_class.__doc__ = (table_class.__doc__ or "") + "\nTable definition:\n\n" + table_class.definition

        # fill values in Lookup tables from their contents property
        if isinstance(instance, Lookup) and hasattr(instance, 'contents') and is_declared:
            contents = list(instance.contents)
            if len(contents) > len(instance):
                if instance.heading.has_autoincrement:
                    warnings.warn(
                        "Contents has changed but cannot be inserted because "
                        "{table} has autoincrement.".format(table=instance.name)
                    )
                else:
                    instance.insert(contents, skip_duplicates=True)

    @property
    def log(self):
        self._assert_exists()
        if self._log is None:
            self._log = Log(self.connection, self.database)
        return self._log

    def __repr__(self):
        return 'Schema `{name}`\n'.format(name=self.database)

    @property
    def size_on_disk(self):
        """
        :return: size of the entire schema in bytes
        """
        self._assert_exists()
        return int(self.connection.query(
            """
            SELECT SUM(data_length + index_length)
            FROM information_schema.tables WHERE table_schema='{db}'
            """.format(db=self.database)).fetchone()[0])

    def spawn_missing_classes(self, context=None):
        """
        Creates the appropriate python user relation classes from tables in the schema and places them
        in the context.
        :param context: alternative context to place the missing classes into, e.g. locals()
        """
        self._assert_exists()
        if context is None:
            if self.context is not None:
                context = self.context
            else:
                # if context is missing, use the calling namespace
                frame = inspect.currentframe().f_back
                context = frame.f_locals
                del frame
        tables = [
            row[0] for row in self.connection.query('SHOW TABLES in `%s`' % self.database)
            if lookup_class_name('`{db}`.`{tab}`'.format(db=self.database, tab=row[0]), context, 0) is None]
        master_classes = (
            Settingstable, AutoImported, AutoComputed,
            Lookup, Manual, Imported, Computed,
        )
        part_tables = []
        for table_name in tables:
            class_name = to_camel_case(table_name)
            if class_name not in context:
                try:
                    cls = next(cls for cls in master_classes if re.fullmatch(cls.tier_regexp, table_name))
                except StopIteration:
                    if re.fullmatch(Part.tier_regexp, table_name):
                        part_tables.append(table_name)
                else:
                    # declare and decorate master relation classes
                    context[class_name] = self(type(class_name, (cls,), dict()), context=context)

        # attach parts to masters
        for table_name in part_tables:
            groups = re.fullmatch(Part.tier_regexp, table_name).groupdict()
            class_name = to_camel_case(groups['part'])
            try:
                master_class = context[to_camel_case(groups['master'])]
            except KeyError:
                raise DataJointError('The table %s does not follow DataJoint naming conventions' % table_name)
            part_class = type(class_name, (Part,), dict(definition=...))
            part_class._master = master_class
            self._decorate_table(part_class, context=context, assert_declared=True)
            setattr(master_class, class_name, part_class)

    def drop(self, force=False):
        """
        Drop the associated schema if it exists
        """
        if not self.exists:
            logger.info("Schema named `{database}` does not exist. Doing nothing.".format(
                database=self.database))
        elif (not config['safemode'] or
              force or
              user_choice("Proceed to delete entire schema `%s`?" % self.database, default='no') == 'yes'):
            logger.info("Dropping `{database}`.".format(database=self.database))
            try:
                self.connection.query("DROP DATABASE `{database}`".format(database=self.database))
                logger.info("Schema `{database}` was dropped successfully.".format(database=self.database))
            except AccessError:
                raise AccessError(
                    "An attempt to drop schema `{database}` "
                    "has failed. Check permissions.".format(database=self.database))

    @property
    def exists(self):
        """
        :return: true if the associated schema exists on the server
        """
        if self.database is None:
            raise DataJointError("Schema must be activated first.")
        return bool(self.connection.query(
            "SELECT schema_name "
            "FROM information_schema.schemata "
            "WHERE schema_name = '{database}'".format(database=self.database)).rowcount)

    @property
    def jobs(self):
        """
        schema.jobs provides a view of the job reservation table for the schema
        :return: jobs table
        """
        self._assert_exists()
        if self._jobs is None:
            self._jobs = JobTable(self.connection, self.database)
        return self._jobs

    @property
    def code(self):
        self._assert_exists()
        return self.save()

    def save(self, python_filename=None):
        """
        Generate the code for a module that recreates the schema.
        This method is in preparation for a future release and is not officially supported.
        :return: a string containing the body of a complete Python module defining this schema.
        """
        self._assert_exists()
        module_count = itertools.count()
        # add virtual modules for referenced modules with names vmod0, vmod1, ...
        module_lookup = collections.defaultdict(lambda: 'vmod' + str(next(module_count)))
        db = self.database

        def make_class_definition(table):
            tier = _get_tier(table).__name__
            class_name = table.split('.')[1].strip('`')
            indent = ''
            if tier == 'Part':
                class_name = class_name.split('__')[-1]
                indent += '    '
            class_name = to_camel_case(class_name)

            def replace(s):
                d, tabs = s.group(1), s.group(2)
                return ('' if d == db else (module_lookup[d]+'.')) + '.'.join(
                    to_camel_case(tab) for tab in tabs.lstrip('__').split('__'))

            return ('' if tier == 'Part' else '\n@schema\n') + (
                '{indent}class {class_name}(dj.{tier}):\n'
                '{indent}    definition = """\n'
                '{indent}    {defi}"""').format(
                    class_name=class_name,
                    indent=indent,
                    tier=tier,
                    defi=re.sub(r'`([^`]+)`.`([^`]+)`', replace,
                                FreeTable(self.connection, table).describe(printout=False)
                                ).replace('\n', '\n    ' + indent))

        diagram = Diagram(self)
        body = '\n\n'.join(make_class_definition(table) for table in diagram.topological_sort())
        python_code = '\n\n'.join((
            '"""This module was auto-generated by datajoint from an existing schema"""',
            "import datajoint as dj\n\nschema = dj.Schema('{db}')".format(db=db),
            '\n'.join("{module} = dj.VirtualModule('{module}', '{schema_name}')".format(module=v, schema_name=k)
                      for k, v in module_lookup.items()), body))
        if python_filename is None:
            return python_code
        with open(python_filename, 'wt') as f:
            f.write(python_code)

    def list_tables(self):
        """
        Return a list of all tables in the schema except tables with ~ in first character such
        as ~logs and ~job
        :return: A list of table names from the database schema.
        """
        return [table_name for (table_name,) in self.connection.query("""
            SELECT table_name FROM information_schema.tables
            WHERE table_schema = %s and table_name NOT LIKE '~%%'""", args=(self.database,))]


class VirtualModule(types.ModuleType):
    """
    A virtual module imitates a Python module representing a DataJoint schema from table definitions in the database.
    It declares the schema objects and a class for each table.
    """
    def __init__(self, module_name, schema_name, *, create_schema=False,
                 create_tables=False, connection=None, add_objects=None):
        """
        Creates a python module with the given name from the name of a schema on the server and
        automatically adds classes to it corresponding to the tables in the schema.
        :param module_name: displayed module name
        :param schema_name: name of the database in mysql
        :param create_schema: if True, create the schema on the database server
        :param create_tables: if True, module.schema can be used as the decorator for declaring new
        :param connection: a dj.Connection object to pass into the schema
        :param add_objects: additional objects to add to the module
        :return: the python module containing classes from the schema object and the table classes
        """
        super(VirtualModule, self).__init__(name=module_name)
        _schema = Schema(schema_name, create_schema=create_schema,
                         create_tables=create_tables, connection=connection)
        if add_objects:
            self.__dict__.update(add_objects)
        self.__dict__['schema'] = _schema
        _schema.spawn_missing_classes(context=self.__dict__)


def list_schemas(connection=None):
    """
    :param connection: a dj.Connection object
    :return: list of all accessible schemas on the server
    """
    return [r[0] for r in (connection or conn()).query(
        'SELECT schema_name '
        'FROM information_schema.schemata '
        'WHERE schema_name <> "information_schema"')]



"""
This module contains the Connection class that manages the connection to the database,
 and the `conn` function that provides access to a persistent connection in datajoint.
"""
import warnings
from contextlib import contextmanager
import pymysql as client
import logging
from getpass import getpass
import re
import pathlib

from .settings import config
from . import errors
from .dependencies import Dependencies
from .blob import pack, unpack
from .hash import uuid_from_buffer
from .plugin import connection_plugins

logger = logging.getLogger(__name__)
query_log_max_length = 300


def get_host_hook(host_input):
    if '://' in host_input:
        plugin_name = host_input.split('://')[0]
        try:
            return connection_plugins[plugin_name]['object'].load().get_host(host_input)
        except KeyError:
            raise errors.DataJointError(
                "Connection plugin '{}' not found.".format(plugin_name))
    else:
        return host_input


def connect_host_hook(connection_obj):
    if '://' in connection_obj.conn_info['host_input']:
        plugin_name = connection_obj.conn_info['host_input'].split('://')[0]
        try:
            connection_plugins[plugin_name]['object'].load().connect_host(connection_obj)
        except KeyError:
            raise errors.DataJointError(
                "Connection plugin '{}' not found.".format(plugin_name))
    else:
        connection_obj.connect()


def translate_query_error(client_error, query):
    """
    Take client error and original query and return the corresponding DataJoint exception.
    :param client_error: the exception raised by the client interface
    :param query: sql query with placeholders
    :return: an instance of the corresponding subclass of datajoint.errors.DataJointError
    """
    logger.debug('type: {}, args: {}'.format(type(client_error), client_error.args))

    err, *args = client_error.args

    # Loss of connection errors
    if err in (0, "(0, '')"):
        return errors.LostConnectionError('Server connection lost due to an interface error.', *args)
    if err == 2006:
        return errors.LostConnectionError("Connection timed out", *args)
    if err == 2013:
        return errors.LostConnectionError("Server connection lost", *args)
    # Access errors
    if err in (1044, 1142):
        return errors.AccessError('Insufficient privileges.', args[0],  query)
    # Integrity errors
    if err == 1062:
        return errors.DuplicateError(*args)
    if err == 1451:
        return errors.IntegrityError(*args)
    if err == 1452:
        return errors.IntegrityError(*args)
    # Syntax errors
    if err == 1064:
        return errors.QuerySyntaxError(args[0], query)
    # Existence errors
    if err == 1146:
        return errors.MissingTableError(args[0], query)
    if err == 1364:
        return errors.MissingAttributeError(*args)
    if err == 1054:
        return errors.UnknownAttributeError(*args)
    # all the other errors are re-raised in original form
    return client_error


def conn(host=None, user=None, password=None, *, init_fun=None, reset=False, use_tls=None):
    """
    Returns a persistent connection object to be shared by multiple modules.
    If the connection is not yet established or reset=True, a new connection is set up.
    If connection information is not provided, it is taken from config which takes the
    information from dj_local_conf.json. If the password is not specified in that file
    datajoint prompts for the password.

    :param host: hostname
    :param user: mysql user
    :param password: mysql password
    :param init_fun: initialization function
    :param reset: whether the connection should be reset or not
    :param use_tls: TLS encryption option. Valid options are: True (required),
                    False (required no TLS), None (TLS prefered, default),
                    dict (Manually specify values per
                    https://dev.mysql.com/doc/refman/5.7/en/connection-options.html
                        #encrypted-connection-options).
    """
    if not hasattr(conn, 'connection') or reset:
        host_input = host if host is not None else config['database.host']
        host = get_host_hook(host_input)
        user = user if user is not None else config['database.user']
        password = password if password is not None else config['database.password']
        if user is None:  # pragma: no cover
            user = input("Please enter DataJoint username: ")
        if password is None:  # pragma: no cover
            password = getpass(prompt="Please enter DataJoint password: ")
        init_fun = init_fun if init_fun is not None else config['connection.init_function']
        use_tls = use_tls if use_tls is not None else config['database.use_tls']
        conn.connection = Connection(host, user, password, None, init_fun, use_tls,
                                     host_input=host_input)
    return conn.connection


class EmulatedCursor:
    """acts like a cursor"""
    def __init__(self, data):
        self._data = data
        self._iter = iter(self._data)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._iter)

    def fetchall(self):
        return self._data

    def fetchone(self):
        return next(self._iter)

    @property
    def rowcount(self):
        return len(self._data)


class Connection:
    """
    A dj.Connection object manages a connection to a database server.
    It also catalogues modules, schemas, tables, and their dependencies (foreign keys).

    Most of the parameters below should be set in the local configuration file.

    :param host: host name, may include port number as hostname:port, in which case it overrides the value in port
    :param user: user name
    :param password: password
    :param port: port number
    :param init_fun: connection initialization function (SQL)
    :param use_tls: TLS encryption option
    """

    def __init__(self, host, user, password, port=None, init_fun=None, use_tls=None,
                 host_input=None):
        if ':' in host:
            # the port in the hostname overrides the port argument
            host, port = host.split(':')
            port = int(port)
        elif port is None:
            port = config['database.port']
        self.conn_info = dict(host=host, port=port, user=user, passwd=password)
        if use_tls is not False:
            self.conn_info['ssl'] = use_tls if isinstance(use_tls, dict) else {'ssl': {}}
        self.conn_info['ssl_input'] = use_tls
        self.conn_info['host_input'] = host_input
        self.init_fun = init_fun
        print("Connecting {user}@{host}:{port}".format(**self.conn_info))
        self._conn = None
        self._query_cache = None
        connect_host_hook(self)
        if self.is_connected:
            logger.info("Connected {user}@{host}:{port}".format(**self.conn_info))
            self.connection_id = self.query('SELECT connection_id()').fetchone()[0]
        else:
            raise errors.LostConnectionError('Connection failed.')
        self._in_transaction = False
        self.schemas = dict()
        self.dependencies = Dependencies(self)

    def __eq__(self, other):
        return self.conn_info == other.conn_info

    def __repr__(self):
        connected = "connected" if self.is_connected else "disconnected"
        return "DataJoint connection ({connected}) {user}@{host}:{port}".format(
            connected=connected, **self.conn_info)

    def connect(self):
        """ Connect to the database server."""
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', '.*deprecated.*')
            try:
                self._conn = client.connect(
                    init_command=self.init_fun,
                    sql_mode="NO_ZERO_DATE,NO_ZERO_IN_DATE,ERROR_FOR_DIVISION_BY_ZERO,"
                             "STRICT_ALL_TABLES,NO_ENGINE_SUBSTITUTION",
                    charset=config['connection.charset'],
                    **{k: v for k, v in self.conn_info.items()
                       if k not in ['ssl_input', 'host_input']})
            except client.err.InternalError:
                self._conn = client.connect(
                    init_command=self.init_fun,
                    sql_mode="NO_ZERO_DATE,NO_ZERO_IN_DATE,ERROR_FOR_DIVISION_BY_ZERO,"
                             "STRICT_ALL_TABLES,NO_ENGINE_SUBSTITUTION",
                    charset=config['connection.charset'],
                    **{k: v for k, v in self.conn_info.items()
                       if not(k in ['ssl_input', 'host_input'] or
                              k == 'ssl' and self.conn_info['ssl_input'] is None)})
        self._conn.autocommit(True)

    def set_query_cache(self, query_cache=None):
        """
        When query_cache is not None, the connection switches into the query caching mode, which entails:
        1. Only SELECT queries are allowed.
        2. The results of queries are cached under the path indicated by dj.config['query_cache']
        3. query_cache is a string that differentiates different cache states.
        :param query_cache: a string to initialize the hash for query results
        """
        self._query_cache = query_cache

    def purge_query_cache(self):
        """ Purges all query cache. """
        if 'query_cache' in config and isinstance(config['query_cache'], str) and \
                pathlib.Path(config['query_cache']).is_dir():
            path_iter = pathlib.Path(config['query_cache']).glob('**/*')
            for path in path_iter:
                path.unlink()

    def close(self):
        self._conn.close()

    def register(self, schema):
        self.schemas[schema.database] = schema
        self.dependencies.clear()

    def ping(self):
        """ Ping the connection or raises an exception if the connection is closed. """
        self._conn.ping(reconnect=False)

    @property
    def is_connected(self):
        """ Return true if the object is connected to the database server. """
        try:
            self.ping()
        except:
            return False
        return True

    @staticmethod
    def _execute_query(cursor, query, args, suppress_warnings):
        try:
            with warnings.catch_warnings():
                if suppress_warnings:
                    # suppress all warnings arising from underlying SQL library
                    warnings.simplefilter("ignore")
                cursor.execute(query, args)
        except client.err.Error as err:
            raise translate_query_error(err, query)

    def query(self, query, args=(), *, as_dict=False, suppress_warnings=True, reconnect=None):
        """
        Execute the specified query and return the tuple generator (cursor).
        :param query: SQL query
        :param args: additional arguments for the client.cursor
        :param as_dict: If as_dict is set to True, the returned cursor objects returns
                        query results as dictionary.
        :param suppress_warnings: If True, suppress all warnings arising from underlying query library
        :param reconnect: when None, get from config, when True, attempt to reconnect if disconnected
        """
        # check cache first:
        use_query_cache = bool(self._query_cache)
        if use_query_cache and not re.match(r"\s*(SELECT|SHOW)", query):
            raise errors.DataJointError("Only SELECT query are allowed when query caching is on.")
        if use_query_cache:
            if not config['query_cache']:
                raise errors.DataJointError("Provide filepath dj.config['query_cache'] when using query caching.")
            hash_ = uuid_from_buffer((str(self._query_cache) + re.sub(r'`\$\w+`', '', query)).encode() + pack(args))
            cache_path = pathlib.Path(config['query_cache']) / str(hash_)
            try:
                buffer = cache_path.read_bytes()
            except FileNotFoundError:
                pass   # proceed to query the database
            else:
                return EmulatedCursor(unpack(buffer))

        if reconnect is None:
            reconnect = config['database.reconnect']
        logger.debug("Executing SQL:" + query[:query_log_max_length])
        cursor_class = client.cursors.DictCursor if as_dict else client.cursors.Cursor
        cursor = self._conn.cursor(cursor=cursor_class)
        try:
            self._execute_query(cursor, query, args, suppress_warnings)
        except errors.LostConnectionError:
            if not reconnect:
                raise
            warnings.warn("MySQL server has gone away. Reconnecting to the server.")
            connect_host_hook(self)
            if self._in_transaction:
                self.cancel_transaction()
                raise errors.LostConnectionError("Connection was lost during a transaction.")
            logger.debug("Re-executing")
            cursor = self._conn.cursor(cursor=cursor_class)
            self._execute_query(cursor, query, args, suppress_warnings)

        if use_query_cache:
            data = cursor.fetchall()
            cache_path.write_bytes(pack(data))
            return EmulatedCursor(data)

        return cursor

    def get_user(self):
        """
        :return: the user name and host name provided by the client to the server.
        """
        return self.query('SELECT user()').fetchone()[0]

    # ---------- transaction processing
    @property
    def in_transaction(self):
        """
        :return: True if there is an open transaction.
        """
        self._in_transaction = self._in_transaction and self.is_connected
        return self._in_transaction

    def start_transaction(self):
        """
        Starts a transaction error.
        """
        if self.in_transaction:
            raise errors.DataJointError("Nested connections are not supported.")
        self.query('START TRANSACTION WITH CONSISTENT SNAPSHOT')
        self._in_transaction = True
        logger.info("Transaction started")

    def cancel_transaction(self):
        """
        Cancels the current transaction and rolls back all changes made during the transaction.
        """
        self.query('ROLLBACK')
        self._in_transaction = False
        logger.info("Transaction cancelled. Rolling back ...")

    def commit_transaction(self):
        """
        Commit all changes made during the transaction and close it.

        """
        self.query('COMMIT')
        self._in_transaction = False
        logger.info("Transaction committed and closed.")

    # -------- context manager for transactions
    @property
    @contextmanager
    def transaction(self):
        """
        Context manager for transactions. Opens an transaction and closes it after the with statement.
        If an error is caught during the transaction, the commits are automatically rolled back.
        All errors are raised again.

        Example:
        >>> import datajoint as dj
        >>> with dj.conn().transaction as conn:
        >>>     # transaction is open here
        """
        try:
            self.start_transaction()
            yield self
        except:
            self.cancel_transaction()
            raise
        else:
            self.commit_transaction()



import networkx as nx
import re
import functools
import io
import warnings
import inspect
from .table import Table
from .dependencies import unite_master_parts

try:
    from matplotlib import pyplot as plt
    plot_active = True
except:
    plot_active = False

try:
    from networkx.drawing.nx_pydot import pydot_layout
    diagram_active = True
except:
    diagram_active = False

from .user_tables import Manual, Imported, Computed, Lookup
from .part_table import Part
from .auto_tables import AutoComputed, AutoImported
from .settings_table import Settingstable
from .errors import DataJointError
from .table import lookup_class_name


user_table_classes = (
    Manual, Lookup, Computed, Imported, Part, AutoComputed, AutoImported,
    Settingstable
)


class _AliasNode:
    """
    special class to indicate aliased foreign keys
    """
    pass


def _get_tier(table_name):
    if not table_name.startswith('`'):
        return _AliasNode
    else:
        try:
            return next(tier for tier in user_table_classes
                        if re.fullmatch(tier.tier_regexp, table_name.split('`')[-2]))
        except StopIteration:
            return None


if not diagram_active:
    class Diagram:
        """
        Entity relationship diagram, currently disabled due to the lack of required packages: matplotlib and pygraphviz.

        To enable Diagram feature, please install both matplotlib and pygraphviz. For instructions on how to install
        these two packages, refer to http://docs.datajoint.io/setup/Install-and-connect.html#python and
        http://tutorials.datajoint.io/setting-up/datajoint-python.html
        """

        def __init__(self, *args, **kwargs):
            warnings.warn('Please install matplotlib and pygraphviz libraries to enable the Diagram feature.')

else:
    class Diagram(nx.DiGraph):
        """
        Entity relationship diagram.

        Usage:

        >>>  diag = Diagram(source)

        source can be a base relation object, a base relation class, a schema, or a module that has a schema.

        >>> diag.draw()

        draws the diagram using pyplot

        diag1 + diag2  - combines the two diagrams.
        diag + n   - expands n levels of successors
        diag - n   - expands n levels of predecessors
        Thus dj.Diagram(schema.Table)+1-1 defines the diagram of immediate ancestors and descendants of schema.Table

        Note that diagram + 1 - 1  may differ from diagram - 1 + 1 and so forth.
        Only those tables that are loaded in the connection object are displayed
        """
        def __init__(self, source, context=None):

            if isinstance(source, Diagram):
                # copy constructor
                self.nodes_to_show = set(source.nodes_to_show)
                self.context = source.context
                super().__init__(source)
                return

            # get the caller's context
            if context is None:
                frame = inspect.currentframe().f_back
                self.context = dict(frame.f_globals, **frame.f_locals)
                del frame
            else:
                self.context = context

            # find connection in the source
            try:
                connection = source.connection
            except AttributeError:
                try:
                    connection = source.schema.connection
                except AttributeError:
                    raise DataJointError('Could not find database connection in %s' % repr(source[0]))

            # initialize graph from dependencies
            connection.dependencies.load()
            super().__init__(connection.dependencies)

            # Enumerate nodes from all the items in the list
            self.nodes_to_show = set()
            try:
                self.nodes_to_show.add(source.full_table_name)
            except AttributeError:
                try:
                    database = source.database
                except AttributeError:
                    try:
                        database = source.schema.database
                    except AttributeError:
                        raise DataJointError('Cannot plot Diagram for %s' % repr(source))
                for node in self:
                    if node.startswith('`%s`' % database):
                        self.nodes_to_show.add(node)

        @classmethod
        def from_sequence(cls, sequence):
            """
            The join Diagram for all objects in sequence
            :param sequence: a sequence (e.g. list, tuple)
            :return: Diagram(arg1) + ... + Diagram(argn)
            """
            return functools.reduce(lambda x, y: x+y, map(Diagram, sequence))

        def add_parts(self):
            """
            Adds to the diagram the part tables of tables already included in the diagram
            :return:
            """
            def is_part(part, master):
                """
                :param part:  `database`.`table_name`
                :param master:   `database`.`table_name`
                :return: True if part is part of master.
                """
                part = [s.strip('`') for s in part.split('.')]
                master = [s.strip('`') for s in master.split('.')]
                return master[0] == part[0] and master[1] + '__' == part[1][:len(master[1])+2]

            self = Diagram(self)  # copy
            self.nodes_to_show.update(n for n in self.nodes() if any(is_part(n, m) for m in self.nodes_to_show))
            return self

        def topological_sort(self):
            """ :return:  list of nodes in topological order """
            return unite_master_parts(list(nx.algorithms.dag.topological_sort(
                nx.DiGraph(self).subgraph(self.nodes_to_show))))

        def __add__(self, arg):
            """
            :param arg: either another Diagram or a positive integer.
            :return: Union of the diagrams when arg is another Diagram
                     or an expansion downstream when arg is a positive integer.
            """
            self = Diagram(self)   # copy
            try:
                self.nodes_to_show.update(arg.nodes_to_show)
            except AttributeError:
                try:
                    self.nodes_to_show.add(arg.full_table_name)
                except AttributeError:
                    for i in range(arg):
                        new = nx.algorithms.boundary.node_boundary(self, self.nodes_to_show)
                        if not new:
                            break
                        # add nodes referenced by aliased nodes
                        new.update(nx.algorithms.boundary.node_boundary(self, (a for a in new if a.isdigit())))
                        self.nodes_to_show.update(new)
            return self

        def __sub__(self, arg):
            """
            :param arg: either another Diagram or a positive integer.
            :return: Difference of the diagrams when arg is another Diagram or
                     an expansion upstream when arg is a positive integer.
            """
            self = Diagram(self)   # copy
            try:
                self.nodes_to_show.difference_update(arg.nodes_to_show)
            except AttributeError:
                try:
                    self.nodes_to_show.remove(arg.full_table_name)
                except AttributeError:
                    for i in range(arg):
                        graph = nx.DiGraph(self).reverse()
                        new = nx.algorithms.boundary.node_boundary(graph, self.nodes_to_show)
                        if not new:
                            break
                        # add nodes referenced by aliased nodes
                        new.update(nx.algorithms.boundary.node_boundary(graph, (a for a in new if a.isdigit())))
                        self.nodes_to_show.update(new)
            return self

        def __mul__(self, arg):
            """
            Intersection of two diagrams
            :param arg: another Diagram
            :return: a new Diagram comprising nodes that are present in both operands.
            """
            self = Diagram(self)   # copy
            self.nodes_to_show.intersection_update(arg.nodes_to_show)
            return self

        def _make_graph(self):
            """
            Make the self.graph - a graph object ready for drawing
            """
            # mark "distinguished" tables, i.e. those that introduce new primary key attributes
            for name in self.nodes_to_show:
                foreign_attributes = set(
                    attr for p in self.in_edges(name, data=True) for attr in p[2]['attr_map'] if p[2]['primary'])
                self.nodes[name]['distinguished'] = (
                        'primary_key' in self.nodes[name] and foreign_attributes < self.nodes[name]['primary_key'])
            # include aliased nodes that are sandwiched between two displayed nodes
            gaps = set(nx.algorithms.boundary.node_boundary(self, self.nodes_to_show)).intersection(
                nx.algorithms.boundary.node_boundary(nx.DiGraph(self).reverse(), self.nodes_to_show))
            nodes = self.nodes_to_show.union(a for a in gaps if a.isdigit)
            # construct subgraph and rename nodes to class names
            graph = nx.DiGraph(nx.DiGraph(self).subgraph(nodes))
            nx.set_node_attributes(graph, name='node_type', values={n: _get_tier(n) for n in graph})
            # relabel nodes to class names
            mapping = {node: lookup_class_name(node, self.context) or node for node in graph.nodes()}
            new_names = [mapping.values()]
            if len(new_names) > len(set(new_names)):
                raise DataJointError('Some classes have identical names. The Diagram cannot be plotted.')
            nx.relabel_nodes(graph, mapping, copy=False)
            return graph

        def make_dot(self):

            graph = self._make_graph()
            graph.nodes()

            scale = 1.2   # scaling factor for fonts and boxes
            label_props = {  # http://matplotlib.org/examples/color/named_colors.html
                None: dict(
                    shape='circle', color="#FFFF0040",
                    fontcolor='yellow', fontsize=round(scale*8),
                    size=0.4*scale, fixed=False),
                _AliasNode: dict(
                    shape='circle', color="#FF880080",
                    fontcolor='#FF880080', fontsize=round(scale*0),
                    size=0.05*scale, fixed=True),
                Manual: dict(
                    shape='box', color='greenyellow',
                    fontcolor='black', fontsize=round(scale*10),
                    size=0.4*scale, fixed=False),
                Settingstable: dict(
                    shape='box', color='coral',
                    fontcolor='black', fontsize=round(scale*8),
                    size=0.4*scale, fixed=False),
                Lookup: dict(
                    shape='plaintext', color='azure4',
                    fontcolor='black', fontsize=round(scale*8),
                    size=0.4*scale, fixed=False),
                Computed: dict(
                    shape='ellipse', color='magenta',
                    fontcolor='black', fontsize=round(scale*10),
                    size=0.3*scale, fixed=True),
                Imported: dict(
                    shape='ellipse', color='skyblue',
                    fontcolor='black', fontsize=round(scale*10),
                    size=0.4*scale, fixed=False),
                AutoComputed: dict(
                    shape='ellipse', color='magenta1',
                    fontcolor='black', fontsize=round(scale*10),
                    size=0.3*scale, fixed=True),
                AutoImported: dict(
                    shape='ellipse', color='skyblue1',
                    fontcolor='black', fontsize=round(scale*10),
                    size=0.4*scale, fixed=False),
                Part: dict(
                    shape='plaintext', color='azure4',
                    fontcolor='black', fontsize=round(scale*8),
                    size=0.1*scale, fixed=False)}
            node_props = {node: label_props[d['node_type']] for node, d in dict(graph.nodes(data=True)).items()}

            dot = nx.drawing.nx_pydot.to_pydot(graph)
            for node in dot.get_nodes():
                node.set_shape('circle')
                name = node.get_name().strip('"')
                props = node_props[name]
                node.set_fontsize(props['fontsize'])
                node.set_fontcolor(props['fontcolor'])
                node.set_shape(props['shape'])
                node.set_fontname('arial')
                node.set_fixedsize('shape' if props['fixed'] else False)
                node.set_width(props['size'])
                node.set_height(props['size'])
                if name.split('.')[0] in self.context:
                    cls = eval(name, self.context)
                    assert(issubclass(cls, Table))
                    description = cls().describe(context=self.context, printout=False).split('\n')
                    description = (
                        '-'*30 if q.startswith('---') else q.replace('->', '&#8594;') if '->' in q else q.split(':')[0]
                        for q in description if not q.startswith('#'))
                    node.set_tooltip('&#13;'.join(description))
                node.set_label("<<u>"+name+"</u>>" if node.get('distinguished') == 'True' else name)
                node.set_color(props['color'])
                node.set_style('filled')

            for edge in dot.get_edges():
                # see http://www.graphviz.org/content/attrs
                src = edge.get_source().strip('"')
                dest = edge.get_destination().strip('"')
                props = graph.get_edge_data(src, dest)
                edge.set_color('#00000040')
                edge.set_style('solid' if props['primary'] else 'dashed')
                master_part = graph.nodes[dest]['node_type'] is Part and dest.startswith(src+'.')
                edge.set_weight(3 if master_part else 1)
                edge.set_arrowhead('none')
                edge.set_penwidth(.75 if props['multi'] else 2)

            return dot

        def make_svg(self):
            from IPython.display import SVG
            return SVG(self.make_dot().create_svg())

        def make_png(self):
            return io.BytesIO(self.make_dot().create_png())

        def make_image(self):
            if plot_active:
                return plt.imread(self.make_png())
            else:
                raise DataJointError("pyplot was not imported")

        def _repr_svg_(self):
            return self.make_svg()._repr_svg_()

        def draw(self):
            if plot_active:
                plt.imshow(self.make_image())
                plt.gca().axis('off')
                plt.show()
            else:
                raise DataJointError("pyplot was not imported")

        def save(self, filename, format=None):
            if format is None:
                if filename.lower().endswith('.png'):
                    format = 'png'
                elif filename.lower().endswith('.svg'):
                    format = 'svg'
            if format.lower() == 'png':
                with open(filename, 'wb') as f:
                    f.write(self.make_png().getbuffer().tobytes())
            elif format.lower() == 'svg':
                with open(filename, 'w') as f:
                    f.write(self.make_svg().data)
            else:
                raise DataJointError('Unsupported file format')

        @staticmethod
        def _layout(graph, **kwargs):
            return pydot_layout(graph, prog='dot', **kwargs)



"""
(De)serialization methods for basic datatypes and numpy.ndarrays with provisions for mutual
compatibility with Matlab-based serialization implemented by mYm.
"""

import zlib
from itertools import repeat
import collections
from decimal import Decimal
import datetime
import uuid
import pickle
import numpy as np
from .errors import DataJointError
from .settings import config

try:
    import cloudpickle
    pickle.dumps = cloudpickle.dumps
except ImportError:
    pass


mxClassID = dict((
    # see http://www.mathworks.com/help/techdoc/apiref/mxclassid.html
    ('mxUNKNOWN_CLASS', None),
    ('mxCELL_CLASS', None),
    ('mxSTRUCT_CLASS', None),
    ('mxLOGICAL_CLASS', np.dtype('bool')),
    ('mxCHAR_CLASS', np.dtype('c')),
    ('mxVOID_CLASS', np.dtype('O')),
    ('mxDOUBLE_CLASS', np.dtype('float64')),
    ('mxSINGLE_CLASS', np.dtype('float32')),
    ('mxINT8_CLASS', np.dtype('int8')),
    ('mxUINT8_CLASS', np.dtype('uint8')),
    ('mxINT16_CLASS', np.dtype('int16')),
    ('mxUINT16_CLASS', np.dtype('uint16')),
    ('mxINT32_CLASS', np.dtype('int32')),
    ('mxUINT32_CLASS', np.dtype('uint32')),
    ('mxINT64_CLASS', np.dtype('int64')),
    ('mxUINT64_CLASS', np.dtype('uint64')),
    ('mxFUNCTION_CLASS', None)))

rev_class_id = {dtype: i for i, dtype in enumerate(mxClassID.values())}
dtype_list = list(mxClassID.values())
type_names = list(mxClassID)

compression = {
    b'ZL123\0': zlib.decompress
}

bypass_serialization = False  # runtime setting to bypass blob (en|de)code


def len_u64(obj):
    return np.uint64(len(obj)).tobytes()


def len_u32(obj):
    return np.uint32(len(obj)).tobytes()


class MatCell(np.ndarray):
    """ a numpy ndarray representing a Matlab cell array """
    pass


class MatStruct(np.recarray):
    """ numpy.recarray representing a Matlab struct array """
    pass


class Blob:
    def __init__(self, squeeze=False):
        self._squeeze = squeeze
        self._blob = None
        self._pos = 0
        self.protocol = None

    def set_dj0(self):
        if not config.get('enable_python_native_blobs'):
            raise DataJointError("""v0.12+ python native blobs disabled.
                See also: https://github.com/datajoint/datajoint-python#python-native-blobs""")

        self.protocol = b"dj0\0"  # when using new blob features

    def set_djpickle(self):
        if not config.get('enable_python_pickle_blobs'):
            raise DataJointError('v0.12+ python pickle blobs disabled.')

        self.protocol = b"djp\0"

    def squeeze(self, array, convert_to_scalar=True):
        """
        Simplify the input array - squeeze out all singleton dimensions.
        If convert_to_scalar, then convert zero-dimensional arrays to scalars
        """
        if not self._squeeze:
            return array
        array = array.squeeze()
        return array.item() if array.ndim == 0 and convert_to_scalar else array

    def unpack(self, blob):
        self._blob = blob
        try:
            # decompress
            prefix = next(p for p in compression if self._blob[self._pos:].startswith(p))
        except StopIteration:
            pass  # assume uncompressed but could be unrecognized compression
        else:
            self._pos += len(prefix)
            blob_size = self.read_value('uint64')
            blob = compression[prefix](self._blob[self._pos:])
            assert len(blob) == blob_size
            self._blob = blob
            self._pos = 0
        blob_format = self.read_zero_terminated_string()
        if blob_format in ('mYm', 'dj0', 'djp'):
            return self.read_blob(n_bytes=len(self._blob) - self._pos)

    def read_blob(self, n_bytes=None):
        start = self._pos
        data_structure_code = chr(self.read_value('uint8'))
        try:
            call = {
                # MATLAB-compatible, inherited from original mYm
                "A": self.read_array,         # matlab-compatible numeric arrays and scalars with ndim==0
                "P": self.read_sparse_array,  # matlab sparse array -- not supported yet
                "S": self.read_struct,        # matlab struct array
                "C": self.read_cell_array,    # matlab cell array
                # basic data types
                "\xFF": self.read_none,    # None
                "\x01": self.read_tuple,     # a Sequence (e.g. tuple)
                "\x02": self.read_list,      # a MutableSequence (e.g. list)
                "\x03": self.read_set,       # a Set
                "\x04": self.read_dict,      # a Mapping (e.g. dict)
                "\x05": self.read_string,    # a UTF8-encoded string
                "\x06": self.read_bytes,     # a ByteString
                "\x0a": self.read_int,       # unbounded scalar int
                "\x0b": self.read_bool,      # scalar boolean
                "\x0c": self.read_complex,   # scalar 128-bit complex number
                "\x0d": self.read_float,     # scalar 64-bit float
                "F": self.read_recarray,     # numpy array with fields, including recarrays
                "d": self.read_decimal,      # a decimal
                "t": self.read_datetime,     # date, time, or datetime
                "u": self.read_uuid,         # UUID
                "\x07": self.read_pickle,
            }[data_structure_code]
        except KeyError:
            raise DataJointError('Unknown data structure code "%s". Upgrade datajoint.' % data_structure_code)
        v = call()
        if n_bytes is not None and self._pos - start != n_bytes:
            raise DataJointError('Blob length check failed! Invalid blob')
        return v

    def pack_blob(self, obj):
        # original mYm-based serialization from datajoint-matlab
        if isinstance(obj, MatCell):
            return self.pack_cell_array(obj)
        if isinstance(obj, MatStruct):
            return self.pack_struct(obj)
        if isinstance(obj, np.ndarray) and obj.dtype.fields is None:
            return self.pack_array(obj)

        # blob types in the expanded dj0 blob format
        self.set_dj0()
        if not isinstance(obj, (np.ndarray, np.number)):
            # python built-in data types
            if isinstance(obj, bool):
                return self.pack_bool(obj)
            if isinstance(obj, int):
                return self.pack_int(obj)
            if isinstance(obj, complex):
                return self.pack_complex(obj)
            if isinstance(obj, float):
                return self.pack_float(obj)
        if isinstance(obj, np.ndarray) and obj.dtype.fields:
            return self.pack_recarray(np.array(obj))
        if isinstance(obj, np.number):
            return self.pack_array(np.array(obj))
        if isinstance(obj, (np.bool, np.bool_)):
            return self.pack_array(np.array(obj))
        if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            return self.pack_datetime(obj)
        if isinstance(obj, Decimal):
            return self.pack_decimal(obj)
        if isinstance(obj, uuid.UUID):
            return self.pack_uuid(obj)
        if isinstance(obj, collections.Mapping):
            return self.pack_dict(obj)
        if isinstance(obj, str):
            return self.pack_string(obj)
        if isinstance(obj, collections.ByteString):
            return self.pack_bytes(obj)
        if isinstance(obj, collections.MutableSequence):
            return self.pack_list(obj)
        if isinstance(obj, collections.Sequence):
            return self.pack_tuple(obj)
        if isinstance(obj, collections.Set):
            return self.pack_set(obj)
        if obj is None:
            return self.pack_none()

        self.set_djpickle()
        return self.pack_pickle(obj)
        # raise DataJointError("Packing object of type %s currently not supported!" % type(obj))

    def read_array(self):
        n_dims = int(self.read_value('uint64'))
        shape = self.read_value('uint64', count=n_dims)
        n_elem = np.prod(shape, dtype=int)
        dtype_id, is_complex = self.read_value('uint32', 2)
        dtype = dtype_list[dtype_id]

        if type_names[dtype_id] == 'mxVOID_CLASS':
            data = np.array(
                list(self.read_blob(self.read_value()) for _ in range(n_elem)), dtype=np.dtype('O'))
        elif type_names[dtype_id] == 'mxCHAR_CLASS':
            # compensate for MATLAB packing of char arrays
            data = self.read_value(dtype, count=2 * n_elem)
            data = data[::2].astype('U1')
            if n_dims == 2 and shape[0] == 1 or n_dims == 1:
                compact = data.squeeze()
                data = compact if compact.shape == () else np.array(''.join(data.squeeze()))
                shape = (1,)
        else:
            data = self.read_value(dtype, count=n_elem)
            if is_complex:
                data = data + 1j * self.read_value(dtype, count=n_elem)
        return self.squeeze(data.reshape(shape, order="F"))

    def pack_array(self, array):
        """
        Serialize an np.ndarray into bytes.  Scalars are encoded with ndim=0.
        """
        blob = b"A" + np.uint64(array.ndim).tobytes() + np.array(array.shape, dtype=np.uint64).tobytes()
        is_complex = np.iscomplexobj(array)
        if is_complex:
            array, imaginary = np.real(array), np.imag(array)
        type_id = (rev_class_id[array.dtype] if array.dtype.char != 'U'
                   else rev_class_id[np.dtype('O')])
        if dtype_list[type_id] is None:
            raise DataJointError("Type %s is ambiguous or unknown" % array.dtype)

        blob += np.array([type_id, is_complex], dtype=np.uint32).tobytes()
        if type_names[type_id] == 'mxVOID_CLASS':  # array of dtype('O')
            blob += b"".join(len_u64(it) + it for it in (self.pack_blob(e) for e in array.flatten(order="F")))
            self.set_dj0()  # not supported by original mym
        elif type_names[type_id] == 'mxCHAR_CLASS':  # array of dtype('c')
            blob += array.view(np.uint8).astype(np.uint16).tobytes()  # convert to 16-bit chars for MATLAB
        else:  # numeric arrays
            if array.ndim == 0:  # not supported by original mym
                self.set_dj0()
            blob += array.tobytes(order="F")
            if is_complex:
                blob += imaginary.tobytes(order="F")
        return blob

    def read_recarray(self):
        """
        Serialize an np.ndarray with fields, including recarrays
        """
        n_fields = self.read_value('uint32')
        if not n_fields:
            return np.array(None)  # empty array
        field_names = [self.read_zero_terminated_string() for _ in range(n_fields)]
        arrays = [self.read_blob() for _ in range(n_fields)]
        rec = np.empty(arrays[0].shape, np.dtype([(f, t.dtype) for f, t in zip(field_names, arrays)]))
        for f, t in zip(field_names, arrays):
            rec[f] = t
        return rec.view(np.recarray)

    def pack_recarray(self, array):
        """ Serialize a Matlab struct array """
        return (b"F" + len_u32(array.dtype) +  # number of fields
                '\0'.join(array.dtype.names).encode() + b"\0" +  # field names
                b"".join(self.pack_recarray(array[f]) if array[f].dtype.fields else self.pack_array(array[f])
                         for f in array.dtype.names))

    def read_sparse_array(self):
        raise DataJointError('datajoint-python does not yet support sparse arrays. Issue (#590)')

    def read_int(self):
        return int.from_bytes(self.read_binary(self.read_value('uint16')), byteorder='little', signed=True)

    @staticmethod
    def pack_int(v):
        n_bytes = v.bit_length() // 8 + 1
        assert 0 < n_bytes <= 0xFFFF, 'Integers are limited to 65535 bytes'
        return b"\x0a" + np.uint16(n_bytes).tobytes() + v.to_bytes(n_bytes, byteorder='little', signed=True)

    def read_bool(self):
        return bool(self.read_value('bool'))

    @staticmethod
    def pack_bool(v):
        return b"\x0b" + np.array(v, dtype='bool').tobytes()

    def read_complex(self):
        return complex(self.read_value('complex128'))

    @staticmethod
    def pack_complex(v):
        return b"\x0c" + np.array(v, dtype='complex128').tobytes()

    def read_float(self):
        return float(self.read_value('float64'))

    @staticmethod
    def pack_float(v):
        return b"\x0d" + np.array(v, dtype='float64').tobytes()

    def read_decimal(self):
        return Decimal(self.read_string())

    @staticmethod
    def pack_decimal(d):
        s = str(d)
        return b"d" + len_u64(s) + s.encode()

    def read_string(self):
        return self.read_binary(self.read_value()).decode()

    @staticmethod
    def pack_string(s):
        blob = s.encode()
        return b"\5" + len_u64(blob) + blob

    def read_bytes(self):
        return self.read_binary(self.read_value())

    @staticmethod
    def pack_bytes(s):
        return b"\6" + len_u64(s) + s

    def read_pickle(self):
        s = self.read_binary(self.read_value())
        return pickle.loads(s)

    @staticmethod
    def pack_pickle(obj):
        s = pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)
        return b"\7" + len_u64(s) + s

    def read_none(self):
        pass

    @staticmethod
    def pack_none():
        return b"\xFF"

    def read_tuple(self):
        return tuple(self.read_blob(self.read_value()) for _ in range(self.read_value()))

    def pack_tuple(self, t):
        return b"\1" + len_u64(t) + b"".join(
            len_u64(it) + it for it in (self.pack_blob(i) for i in t))

    def read_list(self):
        return list(self.read_blob(self.read_value()) for _ in range(self.read_value()))

    def pack_list(self, t):
        return b"\2" + len_u64(t) + b"".join(
            len_u64(it) + it for it in (self.pack_blob(i) for i in t))

    def read_set(self):
        return set(self.read_blob(self.read_value()) for _ in range(self.read_value()))

    def pack_set(self, t):
        return b"\3" + len_u64(t) + b"".join(
            len_u64(it) + it for it in (self.pack_blob(i) for i in t))

    def read_dict(self):
        return dict((self.read_blob(self.read_value()), self.read_blob(self.read_value()))
                    for _ in range(self.read_value()))

    def pack_dict(self, d):
        return b"\4" + len_u64(d) + b"".join(
            b"".join((len_u64(it) + it) for it in packed)
            for packed in (map(self.pack_blob, pair) for pair in d.items()))

    def read_struct(self):
        """deserialize matlab stuct"""
        n_dims = self.read_value()
        shape = self.read_value(count=n_dims)
        n_elem = np.prod(shape, dtype=int)
        n_fields = self.read_value('uint32')
        if not n_fields:
            return np.array(None)  # empty array
        field_names = [self.read_zero_terminated_string() for _ in range(n_fields)]
        raw_data = [
            tuple(self.read_blob(n_bytes=int(self.read_value('uint64'))) for _ in range(n_fields))
            for __ in range(n_elem)]
        data = np.array(raw_data, dtype=list(zip(field_names, repeat(np.object))))
        return self.squeeze(data.reshape(shape, order="F"), convert_to_scalar=False).view(MatStruct)

    def pack_struct(self, array):
        """ Serialize a Matlab struct array """
        return (b"S" + np.array((array.ndim,) + array.shape, dtype=np.uint64).tobytes() +  # dimensionality
                len_u32(array.dtype.names) +  # number of fields
                "\0".join(array.dtype.names).encode() + b"\0" +  # field names
                b"".join(len_u64(it) + it for it in (
                    self.pack_blob(e) for rec in array.flatten(order="F") for e in rec)))  # values

    def read_cell_array(self):
        """ deserialize MATLAB cell array """
        n_dims = self.read_value()
        shape = self.read_value(count=n_dims)
        n_elem = int(np.prod(shape))
        result = [self.read_blob(n_bytes=self.read_value()) for _ in range(n_elem)]
        return (self.squeeze(np.array(result).reshape(shape, order="F"), convert_to_scalar=False)).view(MatCell)

    def pack_cell_array(self, array):
        return (b"C" + np.array((array.ndim,) + array.shape, dtype=np.uint64).tobytes() +
                b"".join(len_u64(it) + it for it in (self.pack_blob(e) for e in array.flatten(order="F"))))

    def read_datetime(self):
        """ deserialize datetime.date, .time, or .datetime """
        date, time = self.read_value('int32'), self.read_value('int64')
        date = datetime.date(
            year=date // 10000,
            month=(date // 100) % 100,
            day=date % 100) if date >= 0 else None
        time = datetime.time(
            hour=(time // 10000000000) % 100,
            minute=(time // 100000000) % 100,
            second=(time // 1000000) % 100,
            microsecond=time % 1000000) if time >= 0 else None
        return time and date and datetime.datetime.combine(date, time) or time or date

    @staticmethod
    def pack_datetime(d):
        if isinstance(d, datetime.datetime):
            date, time = d.date(), d.time()
        elif isinstance(d, datetime.date):
            date, time = d, None
        else:
            date, time = None, d
        return b"t" + (
            np.int32(-1 if date is None else (date.year*100 + date.month)*100 + date.day).tobytes() +
            np.int64(-1 if time is None else
                     ((time.hour*100 + time.minute)*100 + time.second)*1000000 + time.microsecond).tobytes())

    def read_uuid(self):
        q = self.read_binary(16)
        return uuid.UUID(bytes=q)

    @staticmethod
    def pack_uuid(obj):
        return b"u" + obj.bytes

    def read_zero_terminated_string(self):
        target = self._blob.find(b'\0', self._pos)
        data = self._blob[self._pos:target].decode()
        self._pos = target + 1
        return data

    def read_value(self, dtype='uint64', count=1):
        data = np.frombuffer(self._blob, dtype=dtype, count=count, offset=self._pos)
        self._pos += data.dtype.itemsize * data.size
        return data[0] if count == 1 else data

    def read_binary(self, size):
        self._pos += int(size)
        return self._blob[self._pos-int(size):self._pos]

    def pack(self, obj, compress):
        self.protocol = b"mYm\0"  # will be replaced with dj0 if new features are used
        blob = self.pack_blob(obj)  # this may reset the protocol and must precede protocol evaluation
        blob = self.protocol + blob
        if compress and len(blob) > 1000:
            compressed = b'ZL123\0' + len_u64(blob) + zlib.compress(blob)
            if len(compressed) < len(blob):
                blob = compressed
        return blob


def pack(obj, compress=True):
    if bypass_serialization:
        # provide a way to move blobs quickly without de/serialization
        assert isinstance(obj, bytes) and obj.startswith((b'ZL123\0', b'mYm\0', b'dj0\0', b'djp\0'))
        return obj
    return Blob().pack(obj, compress=compress)


def unpack(blob, squeeze=False):
    if bypass_serialization:
        # provide a way to move blobs quickly without de/serialization
        assert isinstance(blob, bytes) and blob.startswith((b'ZL123\0', b'mYm\0', b'dj0\0', b'djp\0'))
        return blob
    if blob is not None:
        return Blob(squeeze=squeeze).unpack(blob)



"""
Settings for DataJoint.
"""
from contextlib import contextmanager
import json
import os
import pprint
import logging
import collections
from enum import Enum
from .errors import DataJointError

LOCALCONFIG = 'dj_local_conf.json'
GLOBALCONFIG = '.datajoint_config.json'
# subfolding for external storage in filesystem.
# 2, 2 means that file abcdef is stored as /ab/cd/abcdef
DEFAULT_SUBFOLDING = (2, 2)

validators = collections.defaultdict(lambda: lambda value: True)
validators['database.port'] = lambda a: isinstance(a, int)

Role = Enum(
    'Role', 'manual lookup imported computed job settingstable autocomputed autoimported'
)
role_to_prefix = {
    Role.manual: '',
    Role.lookup: '#',
    Role.imported: '_',
    Role.computed: '__',
    Role.job: '~',
    Role.autocomputed: '_#',
    Role.autoimported: '#_',
    Role.settingstable: '##'
}
prefix_to_role = dict(zip(role_to_prefix.values(), role_to_prefix))

default = dict({
    'database.host': 'localhost',
    'database.password': None,
    'database.user': None,
    'database.port': 3306,
    'database.reconnect': True,
    'connection.init_function': None,
    'connection.charset': '',   # pymysql uses '' as default
    'loglevel': 'INFO',
    'safemode': True,
    'fetch_format': 'array',
    'display.limit': 12,
    'display.width': 14,
    'display.show_tuple_count': True,
    'database.use_tls': None,
    'enable_python_native_blobs': True,  # python-native/dj0 encoding support
    'enable_python_pickle_blobs': True,
    'enable_automakers': True,
    'automaker_warnings': True,
    'tmp_folder': None,
    'backup_context': None,
    'delete_permission': None
})

logger = logging.getLogger(__name__)
log_levels = {
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'CRITICAL': logging.CRITICAL,
    'DEBUG': logging.DEBUG,
    'ERROR': logging.ERROR,
    None:  logging.NOTSET
}


class Config(collections.MutableMapping):

    instance = None

    def __init__(self, *args, **kwargs):
        if not Config.instance:
            Config.instance = Config.__Config(*args, **kwargs)
        else:
            Config.instance._conf.update(dict(*args, **kwargs))

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __getitem__(self, item):
        return self.instance.__getitem__(item)

    def __setitem__(self, item, value):
        self.instance.__setitem__(item, value)

    def __str__(self):
        return pprint.pformat(self.instance._conf, indent=4)

    def __repr__(self):
        return self.__str__()

    def __delitem__(self, key):
        del self.instance._conf[key]

    def __iter__(self):
        return iter(self.instance._conf)

    def __len__(self):
        return len(self.instance._conf)

    def save(self, filename, verbose=False):
        """
        Saves the settings in JSON format to the given file path.
        :param filename: filename of the local JSON settings file.
        :param verbose: report having saved the settings file
        """
        with open(filename, 'w') as fid:
            json.dump(self._conf, fid, indent=4)
        if verbose:
            print('Saved settings in ' + filename)

    def load(self, filename):
        """
        Updates the setting from config file in JSON format.
        :param filename: filename of the local JSON settings file. If None, the local config file is used.
        """
        if filename is None:
            filename = LOCALCONFIG
        with open(filename, 'r') as fid:
            self._conf.update(json.load(fid))

    def save_local(self, verbose=False):
        """
        saves the settings in the local config file
        """
        self.save(LOCALCONFIG, verbose)

    def save_global(self, verbose=False):
        """
        saves the settings in the global config file
        """
        self.save(os.path.expanduser(os.path.join('~', GLOBALCONFIG)), verbose)

    def get_store_spec(self, store):
        """
        find configuration of external stores for blobs and attachments
        """
        try:
            spec = self['stores'][store]
        except KeyError:
            raise DataJointError('Storage {store} is requested but not configured'.format(store=store))

        spec['subfolding'] = spec.get('subfolding', DEFAULT_SUBFOLDING)
        spec_keys = {  # REQUIRED in uppercase and allowed in lowercase
            'file': ('PROTOCOL', 'LOCATION', 'subfolding', 'stage'),
            's3': ('PROTOCOL', 'ENDPOINT', 'BUCKET', 'ACCESS_KEY', 'SECRET_KEY', 'LOCATION', 'secure', 'subfolding', 'stage')}

        try:
            spec_keys = spec_keys[spec.get('protocol', '').lower()]
        except KeyError:
            raise DataJointError(
                'Missing or invalid protocol in dj.config["stores"]["{store}"]'.format(store=store))

        # check that all required keys are present in spec
        try:
            raise DataJointError('dj.config["stores"]["{store}"] is missing "{k}"'.format(
                store=store, k=next(k.lower() for k in spec_keys if k.isupper() and k.lower() not in spec)))
        except StopIteration:
            pass

        # check that only allowed keys are present in spec
        try:
            raise DataJointError('Invalid key "{k}" in dj.config["stores"]["{store}"]'.format(
                store=store, k=next(k for k in spec if k.upper() not in spec_keys and k.lower() not in spec_keys)))
        except StopIteration:
            pass  # no invalid keys

        return spec

    def get_from_context(self, context, key):
        if context is not None and key in context:
            return context[key]
        elif (
            self['backup_context'] is not None
            and key in self['backup_context']
        ):
            return self['backup_context'][key]
        elif context is not None:
            # will raise KeyError
            return context[key]
        else:
            raise KeyError("Key '{}' not in context".format(key))

    def eval_from_context(self, context, string):
        try:
            return eval(string, context)
        except Exception as e:
            if self['backup_context'] is None:
                raise e
            return eval(string, self['backup_context'])

    @contextmanager
    def __call__(self, **kwargs):
        """
        The config object can also be used in a with statement to change the state of the configuration
        temporarily. kwargs to the context manager are the keys into config, where '.' is replaced by a
        double underscore '__'. The context manager yields the changed config object.

        Example:
        >>> import datajoint as dj
        >>> with dj.config(safemode=False, database__host="localhost") as cfg:
        >>>     # do dangerous stuff here
        """

        try:
            backup = self.instance
            self.instance = Config.__Config(self.instance._conf)
            new = {k.replace('__', '.'): v for k, v in kwargs.items()}
            self.instance._conf.update(new)
            yield self
        except:
            self.instance = backup
            raise
        else:
            self.instance = backup

    class __Config:
        """
        Stores datajoint settings. Behaves like a dictionary, but applies validator functions
        when certain keys are set.

        The default parameters are stored in datajoint.settings.default . If a local config file
        exists, the settings specified in this file override the default settings.
        """

        def __init__(self, *args, **kwargs):
            self._conf = dict(default)
            self._conf.update(dict(*args, **kwargs))  # use the free update to set keys

        def __getitem__(self, key):
            return self._conf[key]

        def __setitem__(self, key, value):
            logger.log(logging.INFO, u"Setting {0:s} to {1:s}".format(str(key), str(value)))
            if validators[key](value):
                self._conf[key] = value
            else:
                raise DataJointError(u'Validator for {0:s} did not pass'.format(key))


# Load configuration from file
config = Config()
config_files = (os.path.expanduser(n) for n in (LOCALCONFIG, os.path.join('~', GLOBALCONFIG)))
try:
    config_file = next(n for n in config_files if os.path.exists(n))
except StopIteration:
    pass
else:
    config.load(config_file)

# override login credentials with environment variables
mapping = {k: v for k, v in zip(
    ('database.host', 'database.user', 'database.password',
     'external.aws_access_key_id', 'external.aws_secret_access_key',),
    map(os.getenv, ('DJ_HOST', 'DJ_USER', 'DJ_PASS',
                    'DJ_AWS_ACCESS_KEY_ID', 'DJ_AWS_SECRET_ACCESS_KEY',)))
           if v is not None}
config.update(mapping)

logger.setLevel(log_levels[config['loglevel']])



"""General-purpose utilities"""

import re
from pathlib import Path
import shutil
from .errors import DataJointError


class ClassProperty:
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


def user_choice(prompt, choices=("yes", "no"), default=None):
    """
    Prompts the user for confirmation.  The default value, if any, is capitalized.
    :param prompt: Information to display to the user.
    :param choices: an iterable of possible choices.
    :param default: default choice
    :return: the user's choice
    """
    assert default is None or default in choices
    choice_list = ', '.join((choice.title() if choice == default else choice for choice in choices))
    response = None
    while response not in choices:
        response = input(prompt + ' [' + choice_list + ']: ')
        response = response.lower() if response else default
    return response


def to_camel_case(s):
    """
    Convert names with under score (_) separation into camel case names.
    :param s: string in under_score notation
    :returns: string in CamelCase notation
    Example:
    >>> to_camel_case("table_name")  # returns "TableName"
    """

    def to_upper(match):
        return match.group(0)[-1].upper()

    return re.sub(r'(^|[_\W])+[a-zA-Z]', to_upper, s)


def from_camel_case(s):
    """
    Convert names in camel case into underscore (_) separated names
    :param s: string in CamelCase notation
    :returns: string in under_score notation
    Example:
    >>> from_camel_case("TableName") # yields "table_name"
    """

    def convert(match):
        return ('_' if match.groups()[0] else '') + match.group(0).lower()

    if not re.match(r'[A-Z][a-zA-Z0-9]*', s):
        raise DataJointError(
            'ClassName must be alphanumeric in CamelCase, begin with a capital letter')
    return re.sub(r'(\B[A-Z])|(\b[A-Z])', convert, s)


def safe_write(filepath, blob):
    """
    A two-step write.
    :param filename: full path
    :param blob: binary data
    """
    filepath = Path(filepath)
    if not filepath.is_file():
        filepath.parent.mkdir(parents=True, exist_ok=True)
        temp_file = filepath.with_suffix(filepath.suffix + '.saving')
        temp_file.write_bytes(blob)
        temp_file.rename(filepath)


def safe_copy(src, dest, overwrite=False):
    """
    Copy the contents of src file into dest file as a two-step process. Skip if dest exists already
    """
    src, dest = Path(src), Path(dest)
    if not (dest.exists() and src.samefile(dest)) and (overwrite or not dest.is_file()):
        dest.parent.mkdir(parents=True, exist_ok=True)
        temp_file = dest.with_suffix(dest.suffix + '.copying')
        shutil.copyfile(str(src), str(temp_file))
        temp_file.rename(dest)


def parse_sql(filepath):
    """
    yield SQL statements from an SQL file
    """
    delimiter = ';'
    statement = []
    with Path(filepath).open('rt') as f:
        for line in f:
            line = line.strip()
            if not line.startswith('--') and len(line) > 1:
                if line.startswith('delimiter'):
                    delimiter = line.split()[1]
                else:
                    statement.append(line)
                    if line.endswith(delimiter):
                        yield ' '.join(statement)
                        statement = []



""" methods for generating previews of query expression results in python command line and Jupyter """

from .settings import config


def preview(query_expression, limit, width):
    heading = query_expression.heading
    rel = query_expression.proj(*heading.non_blobs)
    if limit is None:
        limit = config['display.limit']
    if width is None:
        width = config['display.width']
    tuples = rel.fetch(limit=limit + 1, format="array")
    has_more = len(tuples) > limit
    tuples = tuples[:limit]
    columns = heading.names
    widths = {f: min(max([len(f)] +
                         [len(str(e)) for e in tuples[f]] if f in tuples.dtype.names else [len('=BLOB=')]) + 4, width) for f
              in columns}
    templates = {f: '%%-%d.%ds' % (widths[f], widths[f]) for f in columns}
    return (
            ' '.join([templates[f] % ('*' + f if f in rel.primary_key else f) for f in columns]) + '\n' +
            ' '.join(['+' + '-' * (widths[column] - 2) + '+' for column in columns]) + '\n' +
            '\n'.join(' '.join(templates[f] % (tup[f] if f in tup.dtype.names else '=BLOB=')
                               for f in columns) for tup in tuples) +
            ('\n   ...\n' if has_more else '\n') +
            (' (Total: %d)\n' % len(rel) if config['display.show_tuple_count'] else ''))


def repr_html(query_expression):
    heading = query_expression.heading
    rel = query_expression.proj(*heading.non_blobs)
    info = heading.table_status
    tuples = rel.fetch(limit=config['display.limit'] + 1, format='array')
    has_more = len(tuples) > config['display.limit']
    tuples = tuples[0:config['display.limit']]

    css = """
    <style type="text/css">
        .Relation{
            border-collapse:collapse;
        }
        .Relation th{
            background: #A0A0A0; color: #ffffff; padding:4px; border:#f0e0e0 1px solid;
            font-weight: normal; font-family: monospace; font-size: 100%;
        }
        .Relation td{
            padding:4px; border:#f0e0e0 1px solid; font-size:100%;
        }
        .Relation tr:nth-child(odd){
            background: #ffffff;
        }
        .Relation tr:nth-child(even){
            background: #f3f1ff;
        }
        /* Tooltip container */
        .djtooltip {
        }
        /* Tooltip text */
        .djtooltip .djtooltiptext {
            visibility: hidden;
            width: 120px;
            background-color: black;
            color: #fff;
            text-align: center;
            padding: 5px 0;
            border-radius: 6px;
            /* Position the tooltip text - see examples below! */
            position: absolute;
            z-index: 1;
        }
        #primary {
            font-weight: bold;
            color: black;
        }
        #nonprimary {
            font-weight: normal;
            color: white;
        }

        /* Show the tooltip text when you mouse over the tooltip container */
        .djtooltip:hover .djtooltiptext {
            visibility: visible;
        }
    </style>
    """
    head_template = """<div class="djtooltip">
                            <p id="{primary}">{column}</p>
                            <span class="djtooltiptext">{comment}</span>
                        </div>"""
    return """
    {css}
    {title}
        <div style="max-height:1000px;max-width:1500px;overflow:auto;">
        <table border="1" class="Relation">
            <thead> <tr style="text-align: right;"> <th> {head} </th> </tr> </thead>
            <tbody> <tr> {body} </tr> </tbody>
        </table>
        {ellipsis}
        {count}</div>
        """.format(
        css=css,
        title="" if info is None else "<b>%s</b>" % info['comment'],
        head='</th><th>'.join(
            head_template.format(column=c, comment=heading.attributes[c].comment,
                                 primary='primary' if c in query_expression.primary_key else 'nonprimary') for c in
            heading.names),
        ellipsis='<p>...</p>' if has_more else '',
        body='</tr><tr>'.join(
            ['\n'.join(['<td>%s</td>' % (tup[name] if name in tup.dtype.names else '=BLOB=')
                        for name in heading.names])
             for tup in tuples]),
        count=('<p>Total: %d</p>' % len(rel)) if config['display.show_tuple_count'] else '')



"""
Exception classes for the DataJoint library
"""

import os


# --- Unverified Plugin Check ---
class PluginWarning(Exception):
    pass


# --- Top Level ---
class DataJointError(Exception):
    """
    Base class for errors specific to DataJoint internal operation.
    """
    def __init__(self, *args):
        from .plugin import connection_plugins, type_plugins
        self.__cause__ = PluginWarning(
            'Unverified DataJoint plugin detected.') if any([any(
                [not plugins[k]['verified'] for k in plugins])
                for plugins in [connection_plugins, type_plugins]
                if plugins]) else None

    def suggest(self, *args):
        """
        regenerate the exception with additional arguments
        :param args: addition arguments
        :return: a new exception of the same type with the additional arguments
        """
        return self.__class__(*(self.args + args))


# --- Second Level ---
class LostConnectionError(DataJointError):
    """
    Loss of server connection
    """


class QueryError(DataJointError):
    """
    Errors arising from queries to the database
    """


# --- Third Level: QueryErrors ---
class QuerySyntaxError(QueryError):
    """
    Errors arising from incorrect query syntax
    """


class AccessError(QueryError):
    """
    User access error: insufficient privileges.
    """


class MissingTableError(DataJointError):
    """
    Query on a table that has not been declared
    """


class DuplicateError(QueryError):
    """
    An integrity error caused by a duplicate entry into a unique key
    """


class IntegrityError(QueryError):
    """
    An integrity error triggered by foreign key constraints
    """


class UnknownAttributeError(QueryError):
    """
    User requests an attribute name not found in query heading
    """


class MissingAttributeError(QueryError):
    """
    An error arising when a required attribute value is not provided in INSERT
    """


class MissingExternalFile(DataJointError):
    """
    Error raised when an external file managed by DataJoint is no longer accessible
    """


class BucketInaccessible(DataJointError):
    """
    Error raised when a S3 bucket is inaccessible
    """


# environment variables to control availability of experimental features

ADAPTED_TYPE_SWITCH = "DJ_SUPPORT_ADAPTED_TYPES"
FILEPATH_FEATURE_SWITCH = "DJ_SUPPORT_FILEPATH_MANAGEMENT"


def _switch_adapted_types(on):
    """
    Enable (on=True) or disable (on=False) support for AttributeAdapter
    """
    if on:
        os.environ[ADAPTED_TYPE_SWITCH] = "TRUE"
    else:
        del os.environ[ADAPTED_TYPE_SWITCH]


def _support_adapted_types():
    """
    check if support for AttributeAdapter is enabled
    """
    return os.getenv(ADAPTED_TYPE_SWITCH, "FALSE").upper() == "TRUE"


def _switch_filepath_types(on):
    """
    Enable (on=True) or disable (on=False) support for AttributeAdapter
    """
    if on:
        os.environ[FILEPATH_FEATURE_SWITCH] = "TRUE"
    else:
        del os.environ[FILEPATH_FEATURE_SWITCH]


def _support_filepath_types():
    """
    check if support for AttributeAdapter is enabled
    """
    return os.getenv(FILEPATH_FEATURE_SWITCH, "FALSE").upper() == "TRUE"



"""This module defines class dj.AutoPopulate"""
import logging
import datetime
import traceback
import random
import inspect
from tqdm import tqdm
from .expression import QueryExpression, AndList
from .errors import DataJointError, LostConnectionError
import signal
import multiprocessing as mp

# noinspection PyExceptionInherit,PyCallingNonCallable

logger = logging.getLogger(__name__)


def initializer(table):
    """Save pickled copy of (disconnected) table to the current process,
    then reconnect to server. For use by call_make_key()"""
    mp.current_process().table = table
    table.connection.connect() # reconnect

def call_make_key(key):
    """Call current process' table.make_key()"""
    table = mp.current_process().table
    error = table.make_key(key)
    return error


class AutoPopulate:
    """
    AutoPopulate is a mixin class that adds the method populate() to a Relation class.
    Auto-populated relations must inherit from both Relation and AutoPopulate,
    must define the property `key_source`, and must define the callback method `make`.
    """
    _key_source = None
    _allow_insert = False

    @property
    def key_source(self):
        """
        :return: the relation whose primary key values are passed, sequentially, to the
                ``make`` method when populate() is called.
                The default value is the join of the parent relations.
                Users may override to change the granularity or the scope of populate() calls.
        """
        def _rename_attributes(table, props):
            return (table.proj(
                **{attr: ref for attr, ref in props['attr_map'].items() if attr != ref})
                if props['aliased'] else table.proj())

        if self._key_source is None:
            parents = self.target.parents(primary=True, as_objects=True, foreign_key_info=True)
            if not parents:
                raise DataJointError(
                    'A relation must have primary dependencies for auto-populate to work')
            self._key_source = _rename_attributes(*parents[0])
            for q in parents[1:]:
                self._key_source *= _rename_attributes(*q)
        return self._key_source

    def make(self, key):
        """
        Derived classes must implement method `make` that fetches data from tables that are
        above them in the dependency hierarchy, restricting by the given key, computes dependent
        attributes, and inserts the new tuples into self.
        """
        raise NotImplementedError('Subclasses of AutoPopulate must implement the method `make`')

    @property
    def target(self):
        """
        :return: table to be populated.
        In the typical case, dj.AutoPopulate is mixed into a dj.Table class by inheritance and the target is self.
        """
        return self

    def _job_key(self, key):
        """
        :param key:  they key returned for the job from the key source
        :return: the dict to use to generate the job reservation hash
        This method allows subclasses to control the job reservation granularity.
        """
        return key

    def _jobs_to_do(self, restrictions):
        """
        :return: the relation containing the keys to be computed (derived from self.key_source)
        """
        if self.restriction:
            raise DataJointError('Cannot call populate on a restricted table. '
                                 'Instead, pass conditions to populate() as arguments.')
        todo = self.key_source

        # key_source is a QueryExpression subclass -- trigger instantiation
        if inspect.isclass(todo) and issubclass(todo, QueryExpression):
            todo = todo()

        if not isinstance(todo, QueryExpression):
            raise DataJointError('Invalid key_source value')
        # check if target lacks any attributes from the primary key of key_source
        try:
            raise DataJointError(
                'The populate target lacks attribute %s from the primary key of key_source' % next(
                    name for name in todo.heading.primary_key if name not in self.target.heading))
        except StopIteration:
            pass
        return (todo & AndList(restrictions)).proj()

    def populate(self, *restrictions, suppress_errors=False, return_exception_objects=False,
                 reserve_jobs=False, order="original", limit=None, max_calls=None,
                 display_progress=False, multiprocess=False):
        """
        rel.populate() calls rel.make(key) for every primary key in self.key_source
        for which there is not already a tuple in rel.
        :param restrictions: a list of restrictions each restrict (rel.key_source - target.proj())
        :param suppress_errors: if True, do not terminate execution.
        :param return_exception_objects: return error objects instead of just error messages
        :param reserve_jobs: if True, reserve jobs to populate in asynchronous fashion
        :param order: "original"|"reverse"|"random"  - the order of execution
        :param limit: if not None, check at most this many keys
        :param max_calls: if not None, populate at most this many keys
        :param display_progress: if True, report progress_bar
        :param multiprocess: if True, use as many processes as CPU cores, or use the integer
        number of processes specified
        """
        if self.connection.in_transaction:
            raise DataJointError('Populate cannot be called during a transaction.')

        valid_order = ['original', 'reverse', 'random']
        if order not in valid_order:
            raise DataJointError('The order argument must be one of %s' % str(valid_order))
        jobs = self.connection.schemas[self.target.database].jobs if reserve_jobs else None

        self._make_key_kwargs = {'suppress_errors':suppress_errors,
                                 'return_exception_objects':return_exception_objects,
                                 'reserve_jobs':reserve_jobs,
                                 'jobs':jobs,
                                }

        # define and set up signal handler for SIGTERM:
        if reserve_jobs:
            def handler(signum, frame):
                logger.info('Populate terminated by SIGTERM')
                raise SystemExit('SIGTERM received')
            old_handler = signal.signal(signal.SIGTERM, handler)

        keys = (self._jobs_to_do(restrictions) - self.target).fetch("KEY", limit=limit)
        if order == "reverse":
            keys.reverse()
        elif order == "random":
            random.shuffle(keys)

        logger.info('Found %d keys to populate' % len(keys))

        if max_calls is not None:
            keys = keys[:max_calls]
        nkeys = len(keys)

        if multiprocess: # True or int, presumably
            if multiprocess is True:
                nproc = mp.cpu_count()
            else:
                if not isinstance(multiprocess, int):
                    raise DataJointError("multiprocess can be False, True or a positive integer")
                nproc = multiprocess
        else:
            nproc = 1
        nproc = min(nproc, nkeys) # no sense spawning more than can be used
        error_list = []
        if nproc > 1: # spawn multiple processes
            # prepare to pickle self:
            self.connection.close() # disconnect parent process from MySQL server
            del self.connection._conn.ctx # SSLContext is not picklable
            print('*** Spawning pool of %d processes' % nproc)
            # send pickled copy of self to each process,
            # each worker process calls initializer(*initargs) when it starts
            with mp.Pool(nproc, initializer, (self,)) as pool:
                if display_progress:
                    with tqdm(total=nkeys, desc=self.name) as pbar:
                        for error in pool.imap(call_make_key, keys, chunksize=1):
                            if error is not None:
                                error_list.append(error)
                            pbar.update()
                else:
                    for error in pool.imap(call_make_key, keys):
                        if error is not None:
                            error_list.append(error)
            self.connection.connect() # reconnect parent process to MySQL server
        else: # use single process
            for key in (tqdm(keys, desc=self.name) if display_progress else keys):
                error = self.make_key(key)
                if error is not None:
                    error_list.append(error)

        del self._make_key_kwargs # clean up
        # restore original signal handler:
        if reserve_jobs:
            signal.signal(signal.SIGTERM, old_handler)

        if suppress_errors:
            return error_list

    def make_key(self, key):
        make = self._make_tuples if hasattr(self, '_make_tuples') else self.make

        kwargs = self._make_key_kwargs
        suppress_errors = kwargs['suppress_errors']
        return_exception_objects = kwargs['return_exception_objects']
        reserve_jobs = kwargs['reserve_jobs']
        jobs = kwargs['jobs']

        if not reserve_jobs or jobs.reserve(self.target.table_name, self._job_key(key)):
            self.connection.start_transaction()
            if key in self.target: # already populated
                self.connection.cancel_transaction()
                if reserve_jobs:
                    jobs.complete(self.target.table_name, self._job_key(key))
            else:
                logger.info('Populating: ' + str(key))
                self.__class__._allow_insert = True
                try:
                    make(dict(key))
                except (KeyboardInterrupt, SystemExit, Exception) as error:
                    try:
                        self.connection.cancel_transaction()
                    except LostConnectionError:
                        pass
                    error_message = '{exception}{msg}'.format(
                        exception=error.__class__.__name__,
                        msg=': ' + str(error) if str(error) else '')
                    if reserve_jobs:
                        # show error name and error message (if any)
                        jobs.error(
                            self.target.table_name, self._job_key(key),
                            error_message=error_message, error_stack=traceback.format_exc())
                    if not suppress_errors or isinstance(error, SystemExit):
                        raise
                    else:
                        logger.error(error)
                        return (key, error if return_exception_objects else error_message)
                else:
                    self.connection.commit_transaction()
                    if reserve_jobs:
                        jobs.complete(self.target.table_name, self._job_key(key))
                finally:
                    self.__class__._allow_insert = False

    def progress(self, *restrictions, display=True):
        """
        report progress of populating the table
        :return: remaining, total -- tuples to be populated
        """
        todo = self._jobs_to_do(restrictions)
        total = len(todo)
        remaining = len(todo - self.target)
        if display:
            print('%-20s' % self.name,
                  'Completed %d of %d (%2.1f%%)   %s' % (
                      total - remaining, total, 100 - 100 * remaining / (total+1e-12),
                      datetime.datetime.strftime(datetime.datetime.now(), '%Y-%m-%d %H:%M:%S')), flush=True)
        return remaining, total



import numpy as np
from collections import namedtuple, defaultdict
from itertools import chain
import re
import logging
from .errors import DataJointError, _support_filepath_types, FILEPATH_FEATURE_SWITCH
from .declare import UUID_DATA_TYPE, SPECIAL_TYPES, TYPE_PATTERN, EXTERNAL_TYPES, NATIVE_TYPES
from .attribute_adapter import get_adapter, AttributeAdapter
from .settings import config


logger = logging.getLogger(__name__)

default_attribute_properties = dict(
    # these default values are set in computed attributes
    name=None, type='expression', in_key=False, nullable=False, default=None,
    comment='calculated attribute',
    autoincrement=False, numeric=None, string=None, uuid=False, is_blob=False,
    is_attachment=False, is_filepath=False,
    is_external=False, adapter=None,
    store=None, unsupported=False, attribute_expression=None,
    database=None, dtype=object)


class Attribute(namedtuple('_Attribute', default_attribute_properties)):
    """
    Properties of a table column (attribute)
    """
    def todict(self):
        """Convert namedtuple to dict."""
        return dict((name, self[i]) for i, name in enumerate(self._fields))

    @property
    def sql_type(self):
        """
        :return: datatype (as string) in database. In most cases, it is the same as self.type
        """
        if self.uuid:
            return UUID_DATA_TYPE
        elif self.adapter is None:
            return self.type
        else:
            return self.adapter.attribute_type

    @property
    def sql_comment(self):
        """ :return: full comment for the SQL declaration. Includes custom type specification """
        return (':uuid:' if self.uuid else '') + self.comment

    @property
    def sql(self):
        """
        Convert primary key attribute tuple into its SQL CREATE TABLE clause.
        Default values are not reflected.
        This is used for declaring foreign keys in referencing tables
        :return: SQL code for attribute declaration
        """
        return '`{name}` {type} NOT NULL COMMENT "{comment}"'.format(
            name=self.name, type=self.sql_type, comment=self.sql_comment)

    @property
    def original_name(self):
        if self.attribute_expression is None:
            return self.name
        assert self.attribute_expression.startswith('`')
        return self.attribute_expression.strip('`')


class Heading:
    """
    Local class for relations' headings.
    Heading contains the property attributes, which is an dict in which the keys are
    the attribute names and the values are Attributes.
    """

    def __init__(self, attribute_specs=None, table_info=None):
        """
        :param attribute_specs: a list of dicts with the same keys as Attribute
        :param table_info: a dict with information to load the heading from the database
        """
        self.indexes = None
        self.table_info = table_info
        self._table_status = None
        self._attributes = None if attribute_specs is None else dict(
            (q['name'], Attribute(**q)) for q in attribute_specs)

    def __len__(self):
        return 0 if self.attributes is None else len(self.attributes)

    @property
    def table_status(self):
        if self.table_info is None:
            return None
        if self._table_status is None:
            self._init_from_database()
        return self._table_status

    @property
    def attributes(self):
        if self._attributes is None:
            self._init_from_database()   # lazy loading from database
        return self._attributes

    @property
    def names(self):
        return [k for k in self.attributes]

    @property
    def primary_key(self):
        return [k for k, v in self.attributes.items() if v.in_key]

    @property
    def secondary_attributes(self):
        return [k for k, v in self.attributes.items() if not v.in_key]

    @property
    def blobs(self):
        return [k for k, v in self.attributes.items() if v.is_blob]

    @property
    def non_blobs(self):
        return [k for k, v in self.attributes.items() if not v.is_blob and not v.is_attachment and not v.is_filepath]

    @property
    def new_attributes(self):
        return [k for k, v in self.attributes.items() if v.attribute_expression is not None]

    def __getitem__(self, name):
        """shortcut to the attribute"""
        return self.attributes[name]

    def __repr__(self):
        """
        :return:  heading representation in DataJoint declaration format but without foreign key expansion
        """
        in_key = True
        ret = ''
        if self._table_status is not None:
            ret += '# ' + self.table_status['comment'] + '\n'
        for v in self.attributes.values():
            if in_key and not v.in_key:
                ret += '---\n'
                in_key = False
            ret += '%-20s : %-28s # %s\n' % (
                v.name if v.default is None else '%s=%s' % (v.name, v.default),
                '%s%s' % (v.type, 'auto_increment' if v.autoincrement else ''), v.comment)
        return ret

    @property
    def has_autoincrement(self):
        return any(e.autoincrement for e in self.attributes.values())

    @property
    def as_dtype(self):
        """
        represent the heading as a numpy dtype
        """
        return np.dtype(dict(
            names=self.names,
            formats=[v.dtype for v in self.attributes.values()]))

    def as_sql(self, fields, include_aliases=True):
        """
        represent heading as the SQL SELECT clause.
        """
        return ','.join(
            '`%s`' % name if self.attributes[name].attribute_expression is None
            else self.attributes[name].attribute_expression + (' as `%s`' % name if include_aliases else '')
            for name in fields)

    def __iter__(self):
        return iter(self.attributes)

    def _init_from_database(self):
        """ initialize heading from an existing database table. """
        conn, database, table_name, context = (
            self.table_info[k] for k in ('conn', 'database', 'table_name', 'context'))
        info = conn.query('SHOW TABLE STATUS FROM `{database}` WHERE name="{table_name}"'.format(
            table_name=table_name, database=database), as_dict=True).fetchone()
        if info is None:
            if table_name == '~log':
                logger.warning('Could not create the ~log table')
                return
            raise DataJointError('The table `{database}`.`{table_name}` is not defined.'.format(
                table_name=table_name, database=database))
        self._table_status = {k.lower(): v for k, v in info.items()}
        cur = conn.query(
            'SHOW FULL COLUMNS FROM `{table_name}` IN `{database}`'.format(
                table_name=table_name, database=database), as_dict=True)

        attributes = cur.fetchall()

        rename_map = {
            'Field': 'name',
            'Type': 'type',
            'Null': 'nullable',
            'Default': 'default',
            'Key': 'in_key',
            'Comment': 'comment'}

        fields_to_drop = ('Privileges', 'Collation')

        # rename and drop attributes
        attributes = [{rename_map[k] if k in rename_map else k: v
                       for k, v in x.items() if k not in fields_to_drop}
                      for x in attributes]
        numeric_types = {
            ('float', False): np.float64,
            ('float', True): np.float64,
            ('double', False): np.float64,
            ('double', True): np.float64,
            ('tinyint', False): np.int64,
            ('tinyint', True): np.int64,
            ('smallint', False): np.int64,
            ('smallint', True): np.int64,
            ('mediumint', False): np.int64,
            ('mediumint', True): np.int64,
            ('int', False): np.int64,
            ('int', True): np.int64,
            ('bigint', False): np.int64,
            ('bigint', True): np.uint64}

        sql_literals = ['CURRENT_TIMESTAMP']

        # additional attribute properties
        for attr in attributes:

            attr.update(
                in_key=(attr['in_key'] == 'PRI'),
                database=database,
                nullable=attr['nullable'] == 'YES',
                autoincrement=bool(re.search(r'auto_increment', attr['Extra'], flags=re.I)),
                numeric=any(TYPE_PATTERN[t].match(attr['type']) for t in ('DECIMAL', 'INTEGER', 'FLOAT')),
                string=any(TYPE_PATTERN[t].match(attr['type']) for t in ('ENUM', 'TEMPORAL', 'STRING')),
                is_blob=bool(TYPE_PATTERN['INTERNAL_BLOB'].match(attr['type'])),
                uuid=False, is_attachment=False, is_filepath=False, adapter=None,
                store=None, is_external=False, attribute_expression=None)

            if any(TYPE_PATTERN[t].match(attr['type']) for t in ('INTEGER', 'FLOAT')):
                attr['type'] = re.sub(r'\(\d+\)', '', attr['type'], count=1)  # strip size off integers and floats
            attr['unsupported'] = not any((attr['is_blob'], attr['numeric'], attr['numeric']))
            attr.pop('Extra')

            # process custom DataJoint types
            special = re.match(r':(?P<type>[^:]+):(?P<comment>.*)', attr['comment'])
            if special:
                special = special.groupdict()
                attr.update(special)
            # process adapted attribute types
            if special and TYPE_PATTERN['ADAPTED'].match(attr['type']):
                adapter_name = special['type']
                # keeping track of attribute type used in adapter
                true_type = re.match(r'(?P<comment>.*);(?P<type>[^:]+);', attr['comment'])
                true_type = true_type.groupdict()
                attr['comment'] = true_type['comment']
                # if adapter in context use it, if not create artificial adapter
                try:
                    if context is None and config['backup_context'] is None:
                        raise DataJointError('Declaration context is not set')
                    attr.update(adapter=get_adapter(context, adapter_name))
                except DataJointError:
                    # if no adapter, then delay the error until the first invocation
                    # but keep at least sql type
                    class AttrA(AttributeAdapter):
                        attribute_type = true_type['type']
                    attr.update(adapter=AttrA())
                else:
                    attr.update(type=attr['adapter'].attribute_type)
                    if not any(r.match(attr['type']) for r in TYPE_PATTERN.values()):
                        raise DataJointError(
                            "Invalid attribute type '{type}' in adapter object <{adapter_name}>.".format(
                                adapter_name=adapter_name, **attr))
                    special = not any(TYPE_PATTERN[c].match(attr['type']) for c in NATIVE_TYPES)

            if special:
                try:
                    category = next(c for c in SPECIAL_TYPES if TYPE_PATTERN[c].match(attr['type']))
                except StopIteration:
                    if attr['type'].startswith('external'):
                        url = "https://docs.datajoint.io/python/admin/5-blob-config.html" \
                              "#migration-between-datajoint-v0-11-and-v0-12"
                        raise DataJointError('Legacy datatype `{type}`. Migrate your external stores to '
                                             'datajoint 0.12: {url}'.format(url=url, **attr))
                    raise DataJointError('Unknown attribute type `{type}`'.format(**attr))
                if category == 'FILEPATH' and not _support_filepath_types():
                    raise DataJointError("""
                        The filepath data type is disabled until complete validation.
                        To turn it on as experimental feature, set the environment variable
                        {env} = TRUE or upgrade datajoint.
                        """.format(env=FILEPATH_FEATURE_SWITCH))
                attr.update(
                    unsupported=False,
                    is_attachment=category in ('INTERNAL_ATTACH', 'EXTERNAL_ATTACH'),
                    is_filepath=category == 'FILEPATH',
                    # INTERNAL_BLOB is not a custom type but is included for completeness
                    is_blob=category in ('INTERNAL_BLOB', 'EXTERNAL_BLOB'),
                    uuid=category == 'UUID',
                    is_external=category in EXTERNAL_TYPES,
                    store=attr['type'].split('@')[1] if category in EXTERNAL_TYPES else None)

            if attr['in_key'] and any((attr['is_blob'], attr['is_attachment'], attr['is_filepath'])):
                raise DataJointError('Blob, attachment, or filepath attributes are not allowed in the primary key')

            if attr['string'] and attr['default'] is not None and attr['default'] not in sql_literals:
                attr['default'] = '"%s"' % attr['default']

            if attr['nullable']:   # nullable fields always default to null
                attr['default'] = 'null'

            # fill out dtype. All floats and non-nullable integers are turned into specific dtypes
            attr['dtype'] = object
            if attr['numeric'] and not attr['adapter']:
                is_integer = TYPE_PATTERN['INTEGER'].match(attr['type'])
                is_float = TYPE_PATTERN['FLOAT'].match(attr['type'])
                if is_integer and not attr['nullable'] or is_float:
                    is_unsigned = bool(re.match('sunsigned', attr['type'], flags=re.I))
                    t = re.sub(r'\(.*\)', '', attr['type'])    # remove parentheses
                    t = re.sub(r' unsigned$', '', t)   # remove unsigned
                    assert (t, is_unsigned) in numeric_types, 'dtype not found for type %s' % t
                    attr['dtype'] = numeric_types[(t, is_unsigned)]

            if attr['adapter']:
                # restore adapted type name
                attr['type'] = adapter_name

        self._attributes = dict(((q['name'], Attribute(**q)) for q in attributes))

        # Read and tabulate secondary indexes
        keys = defaultdict(dict)
        for item in conn.query('SHOW KEYS FROM `{db}`.`{tab}`'.format(db=database, tab=table_name), as_dict=True):
            if item['Key_name'] != 'PRIMARY':
                keys[item['Key_name']][item['Seq_in_index']] = dict(
                    column=item['Column_name'],
                    unique=(item['Non_unique'] == 0),
                    nullable=item['Null'].lower() == 'yes')
        self.indexes = {
            tuple(item[k]['column'] for k in sorted(item.keys())):
                dict(unique=item[1]['unique'],
                     nullable=any(v['nullable'] for v in item.values()))
            for item in keys.values()}

    def select(self, select_list, rename_map=None, compute_map=None):
        """
        derive a new heading by selecting, renaming, or computing attributes.
        In relational algebra these operators are known as project, rename, and extend.
        :param select_list:  the full list of existing attributes to include
        :param rename_map:  dictionary of renamed attributes: keys=new names, values=old names
        :param compute_map: a direction of computed attributes
        This low-level method performs no error checking.
        """
        rename_map = rename_map or {}
        compute_map = compute_map or {}
        copy_attrs = list()
        for name in self.attributes:
            if name in select_list:
                copy_attrs.append(self.attributes[name].todict())
            copy_attrs.extend((
                dict(self.attributes[old_name].todict(), name=new_name, attribute_expression='`%s`' % old_name)
                for new_name, old_name in rename_map.items() if old_name == name))
        compute_attrs = (dict(default_attribute_properties, name=new_name, attribute_expression=expr)
                         for new_name, expr in compute_map.items())
        return Heading(chain(copy_attrs, compute_attrs))

    def join(self, other):
        """
        Join two headings into a new one.
        It assumes that self and other are headings that share no common dependent attributes.
        """
        return Heading(
            [self.attributes[name].todict() for name in self.primary_key] +
            [other.attributes[name].todict() for name in other.primary_key if name not in self.primary_key] +
            [self.attributes[name].todict() for name in self.secondary_attributes if name not in other.primary_key] +
            [other.attributes[name].todict() for name in other.secondary_attributes if name not in self.primary_key])

    def set_primary_key(self, primary_key):
        """
        Create a new heading with the specified primary key.
        This low-level method performs no error checking.
        """
        return Heading(chain(
            (dict(self.attributes[name].todict(), in_key=True) for name in primary_key),
            (dict(self.attributes[name].todict(), in_key=False) for name in self.names if name not in primary_key)))

    def make_subquery_heading(self):
        """
        Create a new heading with removed attribute sql_expressions.
        Used by subqueries, which resolve the sql_expressions.
        """
        return Heading(dict(v.todict(), attribute_expression=None) for v in self.attributes.values())



"""Defines Subclass of Autopopulate called Automake
"""

import re
import pandas
import collections
import warnings
import numpy as np

from .autopopulate import AutoPopulate
from .expression import AndList
from .utils import ClassProperty
from .errors import DataJointError
from .settings_table import get_settings_table

Sequence = (collections.MutableSequence, tuple, set)


class AutoMake(AutoPopulate):
    """
    AutoMake is a mixin class for autocomputed and autoimported tables.
    It adds a settings table upstream and make method.
    """

    _settings_table_name = None
    _settings_schema = None
    _settings_table = None
    _settings = None
    _verbose = False

    def make_compatible(self, data):
        """function that can be defined for each class to transform data after
        computation.
        """
        return data

    def populate(self, settings_name, *restrictions, **kwargs):
        """
        rel.populate() calls rel.make(key) for every primary key in self.key_source
        for which there is not already a tuple in rel.
        :param settings_name: name of settings to use for autopopulation from SettingsTable.
        :param restrictions: a list of restrictions each restrict (rel.key_source - target.proj())
        :param suppress_errors: if True, do not terminate execution.
        :param return_exception_objects: return error objects instead of just error messages
        :param reserve_jobs: if true, reserves job to populate in asynchronous fashion
        :param order: "original"|"reverse"|"random"  - the order of execution
        :param display_progress: if True, report progress_bar
        :param limit: if not None, checks at most that many keys
        :param max_calls: if not None, populates at max that many keys
        """

        # TODO dealing with restrictions

        setting_restrict = {self.settings_name: settings_name}
        settings_table = self.settings_table & setting_restrict

        if len(settings_table) != 1:
            raise DataJointError(
                'Setting "{}" does not exist'.format(settings_name)
            )
        settings = (settings_table).fetch1()
        settings['fetch_tables'] = (
            settings['fetch_tables'] & AndList(restrictions)
        )
        self._settings = settings
        self._verbose = kwargs.pop('verbose', False)

        return super().populate(
            setting_restrict,
            self._settings['fetch_tables'].proj(),
            *restrictions,
            **kwargs
        )

    def get_entry(self, key):
        """
        Method used within make to get entry/tuple data
        """
        table = self._settings['fetch_tables'] & key

        if 'fetch1' in self._settings['fetch_method']:
            entry = getattr(
                table,
                self._settings['fetch_method']
            )()

        else:
            if len(table) == 0:
                raise DataJointError(
                    'empty joined table for key {0}'.format(key)
                )

            entry = getattr(
                table,
                self._settings['fetch_method']
            )(format='frame').reset_index().to_dict('list')

            for column, value in entry.items():
                if column in self._settings['parse_unique']:
                    # TODO check if unique?
                    entry[column] = value[0]
        return entry

    def prepare_make(self, entry):
        """
        Method used within make to prepare for computation
        """

        args, kwargs = self._create_kwargs(
            entry,
            self._settings['entry_settings'],
            self._settings['global_settings'],
            self._settings['args'],
            self._settings['kwargs']
        )

        func = self._settings['func']

        return func, args, kwargs

    def make(self, key):
        """automated make method
        """

        if self._verbose:
            print("Start autopopulation for key `{0}`".format(key))

        entry = self.get_entry(key)
        func, args, kwargs = self.prepare_make(entry)
        output = func(*args, **kwargs)
        self.insert_output(key, entry, output)

        # verbosity
        if self._verbose:
            print('Populated entry: {key}'.format(key=key))

    def insert_output(self, key, entry, output):
        """
        Method used within make to insert output into table
        """

        if self._settings['assign_output'] is None:
            output = self.make_compatible(output)
        else:
            output = {self._settings['assign_output']: output}

        if output is None:
            warnings.warn('output of function is None for key {0}'.format(key))
            output = {}

        # Test if dict or dataframe, convert to dataframe if necessary
        if isinstance(output, np.recarray):
            output = pandas.DataFrame(output)

        if (
            self.has_parts
            and not isinstance(output, (pandas.DataFrame, dict))
        ):
            raise DataJointError(
                "output must be dataframe or dict for table with part tables."
            )

        elif not self.has_parts and not isinstance(output, dict):
            raise DataJointError(
                "ouput must be dict for table without part tables."
            )

        # settings name - add to output
        output[self.settings_name] = self._settings[self.settings_name]

        # add columns that are missing in the output (only primary keys)
        for column in (set(self.primary_key) & set(entry) - set(output)):
            if (
                'fetch1' in self._settings['fetch_method']
                or column in self._settings['parse_unique']
            ):
                output[column] = entry[column]
            else:
                output[column] = entry[column][0]

        # insert into table and part_table
        if self.has_parts:
            self.insert1p(output, raise_part_missing=False)
        else:
            self.insert1(output)

    @staticmethod
    def _create_kwargs(
        entry, entry_settings, global_settings,
        settings_args, settings_kwargs
    ):
        """create args and kwargs to pass to function
        """
        args = []
        kwargs = global_settings.copy()
        # substitute entry settings
        for kw, arg in entry_settings.items():
            if isinstance(arg, str):
                kwargs[kw] = entry[arg]
            elif isinstance(arg, Sequence):
                kwargs[kw] = type(arg)(
                    [entry[iarg] for iarg in arg]
                )
            elif isinstance(arg, collections.Mapping):
                kwargs[kw] = {
                    key: entry[iarg] for key, iarg in arg.items()
                }
            else:
                raise DataJointError(
                    "argument in entry settings must be "
                    "str, tuple, or list, but is "
                    "{0} for {1}".format(type(arg), kw)
                )

        if settings_args is not None:
            args.extend(kwargs.pop(settings_args))
        if settings_kwargs is not None:
            kw = kwargs.pop(settings_kwargs)
            kwargs.update(kw)
        #
        return args, kwargs

    @property
    def settings_name(self):
        return self.settings_table().settings_name

    @ClassProperty
    def settings_table(cls):
        """return settings table
        """

        if cls._settings_table is None:
            # dynamically assign settings table
            cls._settings_table = get_settings_table(
                cls,
                name=cls._settings_table_name,
                assigned_schema=cls._settings_schema
            )

        return cls._settings_table

    @classmethod
    def set_true_definition(cls):
        """add settings table attribute if not in definition
        """

        settings_table_attribute = '-> {}'.format(cls.settings_table._save_name)

        if isinstance(cls.definition, property):
            pass
        elif settings_table_attribute not in cls.definition:

            definition = re.split(r'\s*\n\s*', cls.definition.strip())

            in_key_index = None

            for line_index, line in enumerate(definition):
                if line.startswith('---') or line.startswith('___'):
                    in_key_index = line_index
                    break

            if in_key_index is None:
                definition.insert(-1, settings_table_attribute)
            else:
                definition.insert(in_key_index, settings_table_attribute)

            cls.definition = '\n'.join(definition)

        return cls

    def primary_parents(self, columns, restrictions=None):
        """returns joined parent tables excluding settings table.
        Uses columns to select what to project and restrictions for
        each table individually.
        """

        joined_primary_parents = None

        if self.target.full_table_name not in self.connection.dependencies:
            self.connection.dependencies.load()

        for freetable in self.target.parents(primary=True, as_objects=True):

            if freetable.full_table_name == self.settings_table.full_table_name:
                continue
            proj_columns = list(set(freetable.heading.names) & set(columns))
            proj_table = freetable.proj(*proj_columns)
            if restrictions is not None:
                proj_table = proj_table & restrictions

            if joined_primary_parents is None:
                joined_primary_parents = proj_table
            else:
                joined_primary_parents = joined_primary_parents * proj_table

        return joined_primary_parents



from itertools import count
import logging
import inspect
import copy
import re
from .settings import config
from .errors import DataJointError
from .fetch import Fetch, Fetch1
from .preview import preview, repr_html
from .condition import AndList, Not, \
    make_condition, assert_join_compatibility, extract_column_names, PromiscuousOperand

logger = logging.getLogger(__name__)


class QueryExpression:
    """
    QueryExpression implements query operators to derive new entity set from its input.
    A QueryExpression object generates a SELECT statement in SQL.
    QueryExpression operators are restrict, join, proj, aggr, and union.

    A QueryExpression object has a support, a restriction (an AndList), and heading.
    Property `heading` (type dj.Heading) contains information about the attributes.
    It is loaded from the database and updated by proj.

    Property `support` is the list of table names or other QueryExpressions to be joined.

    The restriction is applied first without having access to the attributes generated by the projection.
    Then projection is applied by selecting modifying the heading attribute.

    Application of operators does not always lead to the creation of a subquery.
    A subquery is generated when:
        1. A restriction is applied on any computed or renamed attributes
        2. A projection is applied remapping remapped attributes
        3. Subclasses: Join, Aggregation, and Union have additional specific rules.
    """
    _restriction = None
    _restriction_attributes = None
    _left = []  # True for left joins, False for inner joins
    _original_heading = None  # heading before projections

    # subclasses or instantiators must provide values
    _connection = None
    _heading = None
    _support = None

    @property
    def connection(self):
        """ a dj.Connection object """
        assert self._connection is not None
        return self._connection

    @property
    def support(self):
        """ A list of table names or subqueries to from the FROM clause """
        assert self._support is not None
        return self._support

    @property
    def heading(self):
        """ a dj.Heading object, reflects the effects of the projection operator .proj """
        return self._heading

    @property
    def original_heading(self):
        """ a dj.Heading object reflecting the attributes before projection """
        return self._original_heading or self.heading

    @property
    def restriction(self):
        """ a AndList object of restrictions applied to input to produce the result """
        if self._restriction is None:
            self._restriction = AndList()
        return self._restriction

    @property
    def restriction_attributes(self):
        """ the set of attribute names invoked in the WHERE clause """
        if self._restriction_attributes is None:
            self._restriction_attributes = set()
        return self._restriction_attributes

    @property
    def primary_key(self):
        return self.heading.primary_key

    _subquery_alias_count = count()    # count for alias names used in from_clause

    def from_clause(self):
        support = ('(' + src.make_sql() + ') as `_s%x`' % next(
            self._subquery_alias_count) if isinstance(src, QueryExpression) else src for src in self.support)
        clause = next(support)
        for s, left in zip(support, self._left):
            clause += 'NATURAL{left} JOIN {clause}'.format(
                left=" LEFT" if left else "",
                clause=s)
        return clause

    def where_clause(self):
        return '' if not self.restriction else ' WHERE(%s)' % ')AND('.join(
            str(s) for s in self.restriction)

    def make_sql(self, fields=None):
        """
        Make the SQL SELECT statement.
        :param fields: used to explicitly set the select attributes
        """
        distinct = self.heading.names == self.primary_key
        return 'SELECT {distinct}{fields} FROM {from_}{where}'.format(
            distinct="DISTINCT " if distinct else "",
            fields=self.heading.as_sql(fields or self.heading.names),
            from_=self.from_clause(), where=self.where_clause())

    # --------- query operators -----------
    def make_subquery(self):
        """ create a new SELECT statement where self is the FROM clause """
        result = QueryExpression()
        result._connection = self.connection
        result._support = [self]
        result._heading = self.heading.make_subquery_heading()
        return result

    def restrict(self, restriction):
        """
        Produces a new expression with the new restriction applied.
        rel.restrict(restriction)  is equivalent to  rel & restriction.
        rel.restrict(Not(restriction))  is equivalent to  rel - restriction
        The primary key of the result is unaffected.
        Successive restrictions are combined as logical AND:   r & a & b  is equivalent to r & AndList((a, b))
        Any QueryExpression, collection, or sequence other than an AndList are treated as OrLists
        (logical disjunction of conditions)
        Inverse restriction is accomplished by either using the subtraction operator or the Not class.

        The expressions in each row equivalent:

        rel & True                          rel
        rel & False                         the empty entity set
        rel & 'TRUE'                        rel
        rel & 'FALSE'                       the empty entity set
        rel - cond                          rel & Not(cond)
        rel - 'TRUE'                        rel & False
        rel - 'FALSE'                       rel
        rel & AndList((cond1,cond2))        rel & cond1 & cond2
        rel & AndList()                     rel
        rel & [cond1, cond2]                rel & OrList((cond1, cond2))
        rel & []                            rel & False
        rel & None                          rel & False
        rel & any_empty_entity_set          rel & False
        rel - AndList((cond1,cond2))        rel & [Not(cond1), Not(cond2)]
        rel - [cond1, cond2]                rel & Not(cond1) & Not(cond2)
        rel - AndList()                     rel & False
        rel - []                            rel
        rel - None                          rel
        rel - any_empty_entity_set          rel

        When arg is another QueryExpression, the restriction  rel & arg  restricts rel to elements that match at least
        one element in arg (hence arg is treated as an OrList).
        Conversely,  rel - arg  restricts rel to elements that do not match any elements in arg.
        Two elements match when their common attributes have equal values or when they have no common attributes.
        All shared attributes must be in the primary key of either rel or arg or both or an error will be raised.

        QueryExpression.restrict is the only access point that modifies restrictions. All other operators must
        ultimately call restrict()

        :param restriction: a sequence or an array (treated as OR list), another QueryExpression, an SQL condition
        string, or an AndList.
        """
        attributes = set()
        new_condition = make_condition(self, restriction, attributes)
        if new_condition is True:
            return self  # restriction has no effect, return the same object
        # check that all attributes in condition are present in the query
        try:
            raise DataJointError("Attribute `%s` is not found in query." % next(
                attr for attr in attributes if attr not in self.heading.names))
        except StopIteration:
            pass  # all ok
        # If the new condition uses any new attributes, a subquery is required.
        # However, Aggregation's HAVING statement works fine with aliased attributes.
        need_subquery = isinstance(self, Union) or (
                not isinstance(self, Aggregation) and self.heading.new_attributes)
        if need_subquery:
            result = self.make_subquery()
        else:
            result = copy.copy(self)
            result._restriction = AndList(self.restriction)  # copy to preserve the original
        result.restriction.append(new_condition)
        result.restriction_attributes.update(attributes)
        return result

    def restrict_in_place(self, restriction):
        self.__dict__.update(self.restrict(restriction).__dict__)

    def __and__(self, restriction):
        """
        Restriction operator e.g. ``q1 & q2``.
        :return: a restricted copy of the input argument
        See QueryExpression.restrict for more detail.
        """
        return self.restrict(restriction)

    def __xor__(self, restriction):
        """
        Permissive restriction operator ignoring compatibility check  e.g. ``q1 ^ q2``.
        """
        if inspect.isclass(restriction) and issubclass(restriction, QueryExpression):
            restriction = restriction()
        if isinstance(restriction, Not):
            return self.restrict(Not(PromiscuousOperand(restriction.restriction)))
        return self.restrict(PromiscuousOperand(restriction))

    def __sub__(self, restriction):
        """
        Inverted restriction e.g. ``q1 - q2``.
        :return: a restricted copy of the input argument
        See QueryExpression.restrict for more detail.
        """
        return self.restrict(Not(restriction))

    def __neg__(self):
        """
        Convert between restriction and inverted restriction e.g. ``-q1``.
        :return: target restriction
        See QueryExpression.restrict for more detail.
        """
        if isinstance(self, Not):
            return self.restriction
        return Not(self)

    def __mul__(self, other):
        """
        join of query expressions `self` and `other` e.g. ``q1 * q2``.
        """
        return self.join(other)

    def __matmul__(self, other):
        """
        Permissive join of query expressions `self` and `other` ignoring compatibility check
            e.g. ``q1 @ q2``.
        """
        if inspect.isclass(other) and issubclass(other, QueryExpression):
            other = other()  # instantiate
        return self.join(other, semantic_check=False)

    def join(self, other, semantic_check=True, left=False):
        """
        create the joined QueryExpression.
        a * b  is short for A.join(B)
        a @ b  is short for A.join(B, semantic_check=False)
        Additionally, left=True will retain the rows of self, effectively performing a left join.
        """
        # trigger subqueries if joining on renamed attributes
        if isinstance(other, U):
            return other * self
        if inspect.isclass(other) and issubclass(other, QueryExpression):
            other = other()  # instantiate
        if not isinstance(other, QueryExpression):
            raise DataJointError("The argument of join must be a QueryExpression")
        if semantic_check:
            assert_join_compatibility(self, other)
        join_attributes = set(n for n in self.heading.names if n in other.heading.names)
        # needs subquery if FROM class has common attributes with the other's FROM clause
        need_subquery1 = need_subquery2 = bool(
            (set(self.original_heading.names) & set(other.original_heading.names))
            - join_attributes)
        # need subquery if any of the join attributes are derived
        need_subquery1 = need_subquery1 or any(n in self.heading.new_attributes for n in join_attributes)
        need_subquery2 = need_subquery2 or any(n in other.heading.new_attributes for n in join_attributes)
        if need_subquery1:
            self = self.make_subquery()
        if need_subquery2:
            other = other.make_subquery()
        result = QueryExpression()
        result._connection = self.connection
        result._support = self.support + other.support
        result._left = self._left + [left] + other._left
        result._heading = self.heading.join(other.heading)
        result._restriction = AndList(self.restriction)
        result._restriction.append(other.restriction)
        result._original_heading = self.original_heading.join(other.original_heading)
        assert len(result.support) == len(result._left) + 1
        return result

    def __add__(self, other):
        """union e.g. ``q1 + q2``."""
        return Union.create(self, other)

    def proj(self, *attributes, **named_attributes):
        """
        Projection operator.
        :param attributes:  attributes to be included in the result. (The primary key is already included).
        :param named_attributes: new attributes computed or renamed from existing attributes.
        :return: the projected expression.
        Primary key attributes cannot be excluded but may be renamed.
        If the attribute list contains an Ellipsis ..., then all secondary attributes are included too
        Prefixing an attribute name with a dash '-attr' removes the attribute from the list if present.
        Keyword arguments can be used to rename attributes as in name='attr', duplicate them as in name='(attr)', or
        self.proj(...) or self.proj(Ellipsis) -- include all attributes (return self)
        self.proj() -- include only primary key
        self.proj('attr1', 'attr2')  -- include primary key and attributes attr1 and attr2
        self.proj(..., '-attr1', '-attr2')  -- include attributes except attr1 and attr2
        self.proj(name1='attr1') -- include primary key and 'attr1' renamed as name1
        self.proj('attr1', dup='(attr1)') -- include primary key and attribute attr1 twice, with the duplicate 'dup'
        self.proj(k='abs(attr1)') adds the new attribute k with the value computed as an expression (SQL syntax)
        from other attributes available before the projection.
        Each attribute name can only be used once.
        """
        # new attributes in parentheses are included again with the new name without removing original
        duplication_pattern = re.compile(r'\s*\(\s*(?P<name>[a-z][a-z_0-9]*)\s*\)\s*$')
        # attributes without parentheses renamed
        rename_pattern = re.compile(r'\s*(?P<name>[a-z][a-z_0-9]*)\s*$')
        replicate_map = {k: m.group('name')
                         for k, m in ((k, duplication_pattern.match(v)) for k, v in named_attributes.items()) if m}
        rename_map = {k: m.group('name')
                      for k, m in ((k, rename_pattern.match(v)) for k, v in named_attributes.items()) if m}
        compute_map = {k: v for k, v in named_attributes.items()
                       if not duplication_pattern.match(v) and not rename_pattern.match(v)}
        attributes = set(attributes)
        # include primary key
        attributes.update((k for k in self.primary_key if k not in rename_map.values()))
        # include all secondary attributes with Ellipsis
        if Ellipsis in attributes:
            attributes.discard(Ellipsis)
            attributes.update((a for a in self.heading.secondary_attributes
                               if a not in attributes and a not in rename_map.values()))
        try:
            raise DataJointError("%s is not a valid data type for an attribute in .proj" % next(
                a for a in attributes if not isinstance(a, str)))
        except StopIteration:
            pass  # normal case
        # remove excluded attributes, specified as `-attr'
        excluded = set(a for a in attributes if a.strip().startswith('-'))
        attributes.difference_update(excluded)
        excluded = set(a.lstrip('-').strip() for a in excluded)
        attributes.difference_update(excluded)
        try:
            raise DataJointError("Cannot exclude primary key attribute %s", next(
                a for a in excluded if a in self.primary_key))
        except StopIteration:
            pass  # all ok
        # check that all attributes exist in heading
        try:
            raise DataJointError(
                'Attribute `%s` not found.' % next(a for a in attributes if a not in self.heading.names))
        except StopIteration:
            pass  # all ok

        # check that all mentioned names are present in heading
        mentions = attributes.union(replicate_map.values()).union(rename_map.values())
        try:
            raise DataJointError("Attribute '%s' not found." % next(a for a in mentions if not self.heading.names))
        except StopIteration:
            pass  # all ok

        # check that newly created attributes do not clash with any other selected attributes
        try:
            raise DataJointError("Attribute `%s` already exists" % next(
                a for a in rename_map if a in attributes.union(compute_map).union(replicate_map)))
        except StopIteration:
            pass  # all ok
        try:
            raise DataJointError("Attribute `%s` already exists" % next(
                a for a in compute_map if a in attributes.union(rename_map).union(replicate_map)))
        except StopIteration:
            pass  # all ok
        try:
            raise DataJointError("Attribute `%s` already exists" % next(
                a for a in replicate_map if a in attributes.union(rename_map).union(compute_map)))
        except StopIteration:
            pass  # all ok

        # need a subquery if the projection remaps any remapped attributes
        used = set(q for v in compute_map.values() for q in extract_column_names(v))
        used.update(rename_map.values())
        used.update(replicate_map.values())
        used.intersection_update(self.heading.names)
        need_subquery = isinstance(self, Union) or any(
            self.heading[name].attribute_expression is not None for name in used)
        if not need_subquery and self.restriction:
            # need a subquery if the restriction applies to attributes that have been renamed
            need_subquery = any(name in self.restriction_attributes for name in self.heading.new_attributes)

        result = self.make_subquery() if need_subquery else copy.copy(self)
        result._original_heading = result.original_heading
        result._heading = result.heading.select(
            attributes, rename_map=dict(**rename_map, **replicate_map), compute_map=compute_map)
        return result

    def aggr(self, group, *attributes, keep_all_rows=False, **named_attributes):
        """
        Aggregation of the type U('attr1','attr2').aggr(group, computation="QueryExpression")
        has the primary key ('attr1','attr2') and performs aggregation computations for all matching elements of `group`.
        :param group:  The query expression to be aggregated.
        :param keep_all_rows: True=keep all the rows from self. False=keep only rows that match entries in group.
        :param named_attributes: computations of the form new_attribute="sql expression on attributes of group"
        :return: The derived query expression
        """
        if Ellipsis in attributes:
            # expand ellipsis to include only attributes from the left table
            attributes = set(attributes)
            attributes.discard(Ellipsis)
            attributes.update(self.heading.secondary_attributes)
        return Aggregation.create(
            self, group=group, keep_all_rows=keep_all_rows).proj(*attributes, **named_attributes)

    aggregate = aggr  # alias for aggr

    # ---------- Fetch operators --------------------
    @property
    def fetch1(self):
        return Fetch1(self)

    @property
    def fetch(self):
        return Fetch(self)

    def head(self, limit=25, **fetch_kwargs):
        """
        shortcut to fetch the first few entries from query expression.
        Equivalent to fetch(order_by="KEY", limit=25)
        :param limit:  number of entries
        :param fetch_kwargs: kwargs for fetch
        :return: query result
        """
        return self.fetch(order_by="KEY", limit=limit, **fetch_kwargs)

    def tail(self, limit=25, **fetch_kwargs):
        """
        shortcut to fetch the last few entries from query expression.
        Equivalent to fetch(order_by="KEY DESC", limit=25)[::-1]
        :param limit:  number of entries
        :param fetch_kwargs: kwargs for fetch
        :return: query result
        """
        return self.fetch(order_by="KEY DESC", limit=limit, **fetch_kwargs)[::-1]

    def __len__(self):
        """:return: number of elements in the result set e.g. ``len(q1)``."""
        return self.connection.query(
            'SELECT count(DISTINCT {fields}) FROM {from_}{where}'.format(
                fields=self.heading.as_sql(self.primary_key, include_aliases=False),
                from_=self.from_clause(),
                where=self.where_clause())).fetchone()[0]

    def __bool__(self):
        """
        :return: True if the result is not empty. Equivalent to len(self) > 0 but often
            faster e.g. ``bool(q1)``.
        """
        return bool(self.connection.query(
            'SELECT EXISTS(SELECT 1 FROM {from_}{where})'.format(
                from_=self.from_clause(),
                where=self.where_clause())).fetchone()[0])

    def __contains__(self, item):
        """
        returns True if the restriction in item matches any entries in self
            e.g. ``restriction in q1``.
        :param item: any restriction
        (item in query_expression) is equivalent to bool(query_expression & item) but may be
        executed more efficiently.
        """
        return bool(self & item)  # May be optimized e.g. using an EXISTS query

    def __iter__(self):
        """
        returns an iterator-compatible QueryExpression object e.g. ``iter(q1)``.

        :param self: iterator-compatible QueryExpression object
        """
        self._iter_only_key = all(v.in_key for v in self.heading.attributes.values())
        self._iter_keys = self.fetch('KEY')
        return self

    def __next__(self):
        """
        returns the next record on an iterator-compatible QueryExpression object
            e.g. ``next(q1)``.

        :param self: A query expression
        :type self: :class:`QueryExpression`
        :rtype: dict
        """
        try:
            key = self._iter_keys.pop(0)
        except AttributeError:
            # self._iter_keys is missing because __iter__ has not been called.
            raise TypeError("A QueryExpression object is not an iterator. "
                            "Use iter(obj) to create an iterator.")
        except IndexError:
            raise StopIteration
        else:
            if self._iter_only_key:
                return key
            else:
                try:
                    return (self & key).fetch1()
                except DataJointError:
                    # The data may have been deleted since the moment the keys were fetched
                    # -- move on to next entry.
                    return next(self)

    def cursor(self, offset=0, limit=None, order_by=None, as_dict=False):
        """
        See expression.fetch() for input description.
        :return: query cursor
        """
        if offset and limit is None:
            raise DataJointError('limit is required when offset is set')
        sql = self.make_sql()
        if order_by is not None:
            sql += ' ORDER BY ' + ', '.join(order_by)
        if limit is not None:
            sql += ' LIMIT %d' % limit + (' OFFSET %d' % offset if offset else "")
        logger.debug(sql)
        return self.connection.query(sql, as_dict=as_dict)

    def __repr__(self):
        """
        returns the string representation of a QueryExpression object e.g. ``str(q1)``.

        :param self: A query expression
        :type self: :class:`QueryExpression`
        :rtype: str
        """
        return super().__repr__() if config['loglevel'].lower() == 'debug' else self.preview()

    def preview(self, limit=None, width=None):
        """ :return: a string of preview of the contents of the query. """
        return preview(self, limit, width)

    def _repr_html_(self):
        """ :return: HTML to display table in Jupyter notebook. """
        return repr_html(self)


class Aggregation(QueryExpression):
    """
    Aggregation.create(arg, group, comp1='calc1', ..., compn='calcn')  yields an entity set
    with primary key from arg.
    The computed arguments comp1, ..., compn use aggregation calculations on the attributes of
    group or simple projections and calculations on the attributes of arg.
    Aggregation is used QueryExpression.aggr and U.aggr.
    Aggregation is a private class in DataJoint, not exposed to users.
    """
    _left_restrict = None   # the pre-GROUP BY conditions for the WHERE clause
    _subquery_alias_count = count()

    @classmethod
    def create(cls, arg, group, keep_all_rows=False):
        if inspect.isclass(group) and issubclass(group, QueryExpression):
            group = group()   # instantiate if a class
        assert isinstance(group, QueryExpression)
        if keep_all_rows and len(group.support) > 1:
            group = group.make_subquery()  # subquery if left joining a join
        join = arg.join(group, left=keep_all_rows)  # reuse the join logic
        result = cls()
        result._connection = join.connection
        result._heading = join.heading.set_primary_key(arg.primary_key)  # use left operand's primary key
        result._support = join.support
        result._left = join._left
        result._left_restrict = join.restriction  # WHERE clause applied before GROUP BY
        result._grouping_attributes = result.primary_key

        return result

    def where_clause(self):
        return '' if not self._left_restrict else ' WHERE (%s)' % ')AND('.join(
            str(s) for s in self._left_restrict)

    def make_sql(self, fields=None):
        fields = self.heading.as_sql(fields or self.heading.names)
        assert self._grouping_attributes or not self.restriction
        distinct = set(self.heading.names) == set(self.primary_key)
        return 'SELECT {distinct}{fields} FROM {from_}{where}{group_by}'.format(
            distinct="DISTINCT " if distinct else "",
            fields=fields,
            from_=self.from_clause(),
            where=self.where_clause(),
            group_by="" if not self.primary_key else (
                " GROUP BY `%s`" % '`,`'.join(self._grouping_attributes) +
                ("" if not self.restriction else ' HAVING (%s)' % ')AND('.join(self.restriction))))

    def __len__(self):
        return self.connection.query(
            'SELECT count(1) FROM ({subquery}) `${alias:x}`'.format(
                subquery=self.make_sql(),
                alias=next(self._subquery_alias_count))).fetchone()[0]

    def __bool__(self):
        return bool(self.connection.query(
            'SELECT EXISTS({sql})'.format(sql=self.make_sql())))


class Union(QueryExpression):
    """
    Union is the private DataJoint class that implements the union operator.
    """
    @classmethod
    def create(cls, arg1, arg2):
        if inspect.isclass(arg2) and issubclass(arg2, QueryExpression):
            arg2 = arg2()  # instantiate if a class
        if not isinstance(arg2, QueryExpression):
            raise DataJointError(
                "A QueryExpression can only be unioned with another QueryExpression")
        if arg1.connection != arg2.connection:
            raise DataJointError(
                "Cannot operate on QueryExpressions originating from different connections.")
        if set(arg1.primary_key) != set(arg2.primary_key):
            raise DataJointError("The operands of a union must share the same primary key.")
        if set(arg1.heading.secondary_attributes) & set(arg2.heading.secondary_attributes):
            raise DataJointError(
                "The operands of a union must not share any secondary attributes.")
        result = cls()
        result._connection = arg1.connection
        result._heading = arg1.heading.join(arg2.heading)
        result._support = [arg1, arg2]
        return result

    def make_sql(self):
        arg1, arg2 = self._support
        if not arg1.heading.secondary_attributes and not arg2.heading.secondary_attributes:
            # no secondary attributes: use UNION DISTINCT
            fields = arg1.primary_key
            return "({sql1}) UNION ({sql2})".format(
                sql1=arg1.make_sql(fields),
                sql2=arg2.make_sql(fields))
        # with secondary attributes, use union of left join with antijoin
        fields = self.heading.names
        sql1 = arg1.join(arg2, left=True).make_sql(fields)
        sql2 = (arg2 - arg1).proj(
            ..., **{k: 'NULL' for k in arg1.heading.secondary_attributes}).make_sql(fields)
        return "({sql1})  UNION ({sql2})".format(sql1=sql1, sql2=sql2)

    def from_clause(self):
        """ The union does not use a FROM clause """
        assert False

    def where_clause(self):
        """ The union does not use a WHERE clause """
        assert False

    def __len__(self):
        return self.connection.query(
            'SELECT count(1) FROM ({subquery}) `${alias:x}`'.format(
                subquery=self.make_sql(),
                alias=next(QueryExpression._subquery_alias_count))).fetchone()[0]

    def __bool__(self):
        return bool(self.connection.query(
            'SELECT EXISTS({sql})'.format(sql=self.make_sql())))


class U:
    """
    dj.U objects are the universal sets representing all possible values of their attributes.
    dj.U objects cannot be queried on their own but are useful for forming some queries.
    dj.U('attr1', ..., 'attrn') represents the universal set with the primary key attributes attr1 ... attrn.
    The universal set is the set of all possible combinations of values of the attributes.
    Without any attributes, dj.U() represents the set with one element that has no attributes.

    Restriction:

    dj.U can be used to enumerate unique combinations of values of attributes from other expressions.

    The following expression yields all unique combinations of contrast and brightness found in the `stimulus` set:

    >>> dj.U('contrast', 'brightness') & stimulus

    Aggregation:

    In aggregation, dj.U is used for summary calculation over an entire set:

    The following expression yields one element with one attribute `s` containing the total number of elements in
    query expression `expr`:

    >>> dj.U().aggr(expr, n='count(*)')

    The following expressions both yield one element containing the number `n` of distinct values of attribute `attr` in
    query expressio `expr`.

    >>> dj.U().aggr(expr, n='count(distinct attr)')
    >>> dj.U().aggr(dj.U('attr').aggr(expr), 'n=count(*)')

    The following expression yields one element and one attribute `s` containing the sum of values of attribute `attr`
    over entire result set of expression `expr`:

    >>> dj.U().aggr(expr, s='sum(attr)')

    The following expression yields the set of all unique combinations of attributes `attr1`, `attr2` and the number of
    their occurrences in the result set of query expression `expr`.

    >>> dj.U(attr1,attr2).aggr(expr, n='count(*)')

    Joins:

    If expression `expr` has attributes 'attr1' and 'attr2', then expr * dj.U('attr1','attr2') yields the same result
    as `expr` but `attr1` and `attr2` are promoted to the the primary key.  This is useful for producing a join on
    non-primary key attributes.
    For example, if `attr` is in both expr1 and expr2 but not in their primary keys, then expr1 * expr2 will throw
    an error because in most cases, it does not make sense to join on non-primary key attributes and users must first
    rename `attr` in one of the operands.  The expression dj.U('attr') * rel1 * rel2 overrides this constraint.
    """

    def __init__(self, *primary_key):
        self._primary_key = primary_key

    @property
    def primary_key(self):
        return self._primary_key

    def __and__(self, other):
        if inspect.isclass(other) and issubclass(other, QueryExpression):
            other = other()   # instantiate if a class
        if not isinstance(other, QueryExpression):
            raise DataJointError('Set U can only be restricted with a QueryExpression.')
        result = copy.copy(other)
        result._heading = result.heading.set_primary_key(self.primary_key)
        result = result.proj()
        return result

    def join(self, other, left=False):
        """
        Joining U with a query expression has the effect of promoting the attributes of U to the primary key of
        the other query expression.
        :param other: the other query expression to join with.
        :param left: ignored. dj.U always acts as if left=False
        :return: a copy of the other query expression with the primary key extended.
        """
        if inspect.isclass(other) and issubclass(other, QueryExpression):
            other = other()   # instantiate if a class
        if not isinstance(other, QueryExpression):
            raise DataJointError('Set U can only be joined with a QueryExpression.')
        try:
            raise DataJointError(
                'Attribute `%s` not found' % next(k for k in self.primary_key if k not in other.heading.names))
        except StopIteration:
            pass  # all ok
        result = copy.copy(other)
        result._heading = result.heading.set_primary_key(
            other.primary_key + [k for k in self.primary_key if k not in other.primary_key])
        return result

    def __mul__(self, other):
        """ shorthand for join """
        return self.join(other)

    def aggr(self, group, **named_attributes):
        """
        Aggregation of the type U('attr1','attr2').aggr(group, computation="QueryExpression")
        has the primary key ('attr1','attr2') and performs aggregation computations for all matching elements of `group`.
        :param group:  The query expression to be aggregated.
        :param named_attributes: computations of the form new_attribute="sql expression on attributes of group"
        :return: The derived query expression
        """
        if named_attributes.get('keep_all_rows', False):
            raise DataJointError(
                'Cannot set keep_all_rows=True when aggregating on a universal set.')
        return Aggregation.create(self, group=group,  keep_all_rows=False).proj(**named_attributes)

    aggregate = aggr  # alias for aggr



import networkx as nx
import itertools
import re
from collections import defaultdict
from .errors import DataJointError


def unite_master_parts(lst):
    """
    re-order a list of table names so that part tables immediately follow their master tables without breaking
    the topological order.
    Without this correction, a simple topological sort may insert other descendants between master and parts.
    The input list must be topologically sorted.
    :example:
    unite_master_parts(
        ['`s`.`a`', '`s`.`a__q`', '`s`.`b`', '`s`.`c`', '`s`.`c__q`', '`s`.`b__q`', '`s`.`d`', '`s`.`a__r`']) ->
        ['`s`.`a`', '`s`.`a__q`', '`s`.`a__r`', '`s`.`b`', '`s`.`b__q`', '`s`.`c`', '`s`.`c__q`', '`s`.`d`']
    """
    for i in range(2, len(lst)):
        name = lst[i]
        match = re.match(r'(?P<master>`\w+`.`\w+)__\w+`', name)
        if match:  # name is a part table
            master = match.group('master')
            for j in range(i-1, -1, -1):
                if lst[j] == master + '`' or lst[j].startswith(master + '__'):
                    # move from the ith position to the (j+1)th position
                    lst[j+1:i+1] = [name] + lst[j+1:i]
                    break
            else:
                raise DataJointError("Found a part table {name} without its master table.".format(name=name))
    return lst


class Dependencies(nx.DiGraph):
    """
    The graph of dependencies (foreign keys) between loaded tables.

    Note: the 'connection' argument should normally be supplied;
    Empty use is permitted to facilitate use of networkx algorithms which
    internally create objects with the expectation of empty constructors.
    See also: https://github.com/datajoint/datajoint-python/pull/443
    """
    def __init__(self, connection=None):
        self._conn = connection
        self._node_alias_count = itertools.count()
        self._loaded = False
        super().__init__(self)

    def clear(self):
        self._loaded = False
        super().clear()

    def load(self, force=True):
        """
        Load dependencies for all loaded schemas.
        This method gets called before any operation that requires dependencies: delete, drop, populate, progress.
        """
        # reload from scratch to prevent duplication of renamed edges
        if self._loaded and not force:
            return

        self.clear()

        # load primary key info
        keys = self._conn.query("""
                SELECT
                    concat('`', table_schema, '`.`', table_name, '`') as tab, column_name
                FROM information_schema.key_column_usage
                WHERE table_name not LIKE "~%%" AND table_schema in ('{schemas}') AND constraint_name="PRIMARY"
                """.format(schemas="','".join(self._conn.schemas)))
        pks = defaultdict(set)
        for key in keys:
            pks[key[0]].add(key[1])

        # add nodes to the graph
        for n, pk in pks.items():
            self.add_node(n, primary_key=pk)

        # load foreign keys
        keys = ({k.lower(): v for k, v in elem.items()} for elem in self._conn.query("""
        SELECT constraint_name,
            concat('`', table_schema, '`.`', table_name, '`') as referencing_table,
            concat('`', referenced_table_schema, '`.`',  referenced_table_name, '`') as referenced_table,
            column_name, referenced_column_name
        FROM information_schema.key_column_usage
        WHERE referenced_table_name NOT LIKE "~%%" AND (referenced_table_schema in ('{schemas}') OR
            referenced_table_schema is not NULL AND table_schema in ('{schemas}'))
        """.format(schemas="','".join(self._conn.schemas)), as_dict=True))
        fks = defaultdict(lambda: dict(attr_map=dict()))
        for key in keys:
            d = fks[(key['constraint_name'], key['referencing_table'], key['referenced_table'])]
            d['referencing_table'] = key['referencing_table']
            d['referenced_table'] = key['referenced_table']
            d['attr_map'][key['column_name']] = key['referenced_column_name']

        # add edges to the graph
        for fk in fks.values():
            props = dict(
                primary=set(fk['attr_map']) <= set(pks[fk['referencing_table']]),
                attr_map=fk['attr_map'],
                aliased=any(k != v for k, v in fk['attr_map'].items()),
                multi=set(fk['attr_map']) != set(pks[fk['referencing_table']]))
            if not props['aliased']:
                self.add_edge(fk['referenced_table'], fk['referencing_table'], **props)
            else:
                # for aliased dependencies, add an extra node in the format '1', '2', etc
                alias_node = '%d' % next(self._node_alias_count)
                self.add_node(alias_node)
                self.add_edge(fk['referenced_table'], alias_node, **props)
                self.add_edge(alias_node, fk['referencing_table'], **props)

        if not nx.is_directed_acyclic_graph(self):  # pragma: no cover
            raise DataJointError('DataJoint can only work with acyclic dependencies')
        self._loaded = True

    def parents(self, table_name, primary=None):
        """
        :param table_name: `schema`.`table`
        :param primary: if None, then all parents are returned. If True, then only foreign keys composed of
            primary key attributes are considered.  If False, the only foreign keys including at least one non-primary
            attribute are considered.
        :return: dict of tables referenced by the foreign keys of table
        """
        self.load(force=False)
        return {p[0]: p[2] for p in self.in_edges(table_name, data=True)
                if primary is None or p[2]['primary'] == primary}

    def children(self, table_name, primary=None):
        """
        :param table_name: `schema`.`table`
        :param primary: if None, then all children are returned. If True, then only foreign keys composed of
            primary key attributes are considered.  If False, the only foreign keys including at least one non-primary
            attribute are considered.
        :return: dict of tables referencing the table through foreign keys
        """
        self.load(force=False)
        return {p[1]: p[2] for p in self.out_edges(table_name, data=True)
                if primary is None or p[2]['primary'] == primary}

    def descendants(self, full_table_name):
        """
        :param full_table_name:  In form `schema`.`table_name`
        :return: all dependent tables sorted in topological order.  Self is included.
        """
        self.load(force=False)
        nodes = self.subgraph(
            nx.algorithms.dag.descendants(self, full_table_name))
        return unite_master_parts([full_table_name] + list(
            nx.algorithms.dag.topological_sort(nodes)))

    def ancestors(self, full_table_name):
        """
        :param full_table_name:  In form `schema`.`table_name`
        :return: all dependent tables sorted in topological order.  Self is included.
        """
        self.load(force=False)
        nodes = self.subgraph(
            nx.algorithms.dag.ancestors(self, full_table_name))
        return list(reversed(unite_master_parts(list(
            nx.algorithms.dag.topological_sort(nodes)) + [full_table_name])))



import re
from .errors import DataJointError, _support_adapted_types
from .settings import config
from .plugin import type_plugins


class AttributeAdapter:
    """
    Base class for adapter objects for user-defined attribute types.
    """
    @property
    def attribute_type(self):
        """
        :return: a supported DataJoint attribute type to use; e.g. "longblob", "blob@store"
        """
        raise NotImplementedError('Undefined attribute adapter')

    def get(self, value):
        """
        convert value retrieved from the the attribute in a table into the adapted type
        :param value: value from the database
        :return: object of the adapted type
        """
        raise NotImplementedError('Undefined attribute adapter')

    def put(self, obj):
        """
        convert an object of the adapted type into a value that DataJoint can store in a table attribute
        :param obj: an object of the adapted type
        :return: value to store in the database
        """
        raise NotImplementedError('Undefined attribute adapter')


def get_adapter(context, adapter_name):
    """
    Extract the AttributeAdapter object by its name from the context and validate.
    """
    if not _support_adapted_types():
        raise DataJointError('Support for Adapted Attribute types is disabled.')
    adapter_name = adapter_name.lstrip('<').rstrip('>')
    try:
        try:
            adapter = config.get_from_context(context, adapter_name)
        except KeyError:
            adapter = type_plugins[adapter_name]['object'].load()
    except KeyError:
        raise DataJointError(
            "Attribute adapter '{adapter_name}' is not defined.".format(adapter_name=adapter_name))
    if not isinstance(adapter, AttributeAdapter):
        raise DataJointError(
            "Attribute adapter '{adapter_name}' must be an instance of datajoint.AttributeAdapter".format(
                adapter_name=adapter_name))
    if not isinstance(adapter.attribute_type, str) or not re.match(r'^\w', adapter.attribute_type):
        raise DataJointError("Invalid attribute type {type} in attribute adapter '{adapter_name}'".format(
            type=adapter.attribute_type, adapter_name=adapter_name))
    return adapter
